# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4; coding:utf-8 -*-
#
# Copyright 2021 Syeam Bin Abdullah
#
# This file is part of duplicity.
#
# Duplicity is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# Duplicity is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with duplicity; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

"""
Slate backend.
"""

import json
import os
import shutil
import urllib.request
from pathlib import Path

import requests

import duplicity.backend
from duplicity import log
from duplicity.errors import BackendException


class SlateBackend(duplicity.backend.Backend):
    """
    Backend for Slate
    """

    def __init__(self, parsed_url):
        """
        Initialize the Slate backend.

        :param ParsedUrl parsed_url: The parsed URL.
        :raises BackendException: If SLATE_API_KEY is not set.
        """
        duplicity.backend.Backend.__init__(self, parsed_url)
        log.Debug("loading slate backend...")
        if "SLATE_API_KEY" not in os.environ.keys():
            raise BackendException("""You must set an environment variable SLATE_API_KEY
                as the value of your slate API key""")
        else:
            self.key = os.environ["SLATE_API_KEY"]

        if "SLATE_SSL_VERIFY" not in os.environ.keys():
            self.verify = True
        else:
            if "SLATE_SSL_VERIFY" == "0":
                self.verify = False
            else:
                self.verify = True

        data = json.dumps({"data": {"private": "true"}})
        headers = {
            "Content-Type": "application/json",
            "Authorization": f"Basic {self.key}",
        }

        response = requests.post(
            "https://slate.host/api/v1/get",
            data=data,
            headers=headers,
            verify=self.verify,
        )
        if not response.ok:
            raise BackendException("Slate backend requires a valid API key")

        self.slate_id = parsed_url.geturl().split("/")[-1]

    def _put(self, source_path, remote_filename):
        data = json.dumps({"data": {"private": "true"}})
        headers = {
            "Content-Type": "application/json",
            "Authorization": f"Basic {self.key}",
        }

        log.Debug(f"source_path.name: {str(source_path.name)}")
        log.Debug(f"remote_filename: {remote_filename.decode('utf8')}")
        rem_filename = str(os.fsdecode(remote_filename))

        src = Path(os.fsdecode(source_path.name))
        if str(src.name).startswith("mktemp"):
            log.Debug("copying temp file for upload")
            src = shutil.move(str(src), str(src.with_name(rem_filename)))

        log.Debug("response")
        headers = {"Authorization": f"Basic {self.key}"}
        files = {rem_filename: open(str(src), "rb")}
        log.Debug(f"-------------------FILECHECK: {str(files.keys())}")
        response = requests.post(
            url=f"https://uploads.slate.host/api/public/{self.slate_id}",
            files=files,
            headers=headers,
        )
        log.Debug("response handled")

        if not response.ok:
            raise BackendException(f"An error occurred whilst attempting to upload a file: {response}")
        else:
            log.Debug(f"File successfully uploaded to slate with id:{self.slate_id}")

        if str(src).endswith("difftar.gpg"):
            os.remove(str(src))

    def _list(self):
        # Checks if a specific slate has been selected, otherwise lists all slates
        log.Debug(f"Slate ID: {self.slate_id}")
        data = json.dumps({"data": {"private": "true"}})
        headers = {
            "Content-Type": "application/json",
            "Authorization": f"Basic {self.key}",
        }
        response = requests.post(
            "https://slate.host/api/v1/get",
            data=data,
            headers=headers,
            verify=self.verify,
        )

        if not response.ok:
            raise BackendException("Slate backend requires a valid API key")

        slates = response.json()["slates"]
        # log.Debug("SLATES:\n%s"%(slates))
        file_list = []
        for slate in slates:
            if slate["id"] == self.slate_id:
                files = slate["data"]["objects"]
                for f in files:
                    file_list.append(f["name"])
            else:
                log.Debug(f"Could not find slate with id: {self.slate_id}")

        return file_list

    def _get(self, remote_filename, local_path):
        # Downloads chosen file from IPFS by parsing its cid
        found = False
        data = json.dumps({"data": {"private": "true"}})
        headers = {
            "Content-Type": "application/json",
            "Authorization": f"Basic {self.key}",
        }

        response = requests.post(
            "https://slate.host/api/v1/get",
            data=data,
            headers=headers,
            verify=self.verify,
        )

        slates = response.json()["slates"]
        # file_list = self._list()

        # if remote_filename not in file_list:
        #     raise BackendException(u"The chosen file does not exist in the chosen slate")

        for slate in slates:
            if slate["id"] == self.slate_id:
                found = True
                for obj in slate["data"]["objects"]:
                    if obj["name"] == remote_filename.decode("utf8"):
                        cid = obj["url"].split("/")[-1]
                        break
                    else:
                        raise BackendException(
                            f"The file '{remote_filename.decode('utf8')}' could not be found in the specified slate"
                        )

        if not found:
            raise BackendException(f"A slate with id {self.slate_id} does not exist")

        try:
            urllib.request.urlretrieve(f"http://ipfs.io/ipfs/{cid}", os.fsdecode(local_path.name))
            log.Debug(f"Downloaded file with cid: {cid}")
        except NameError as e:
            raise BackendException("Couldn't download file")


duplicity.backend.register_backend("slate", SlateBackend)



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4; coding:utf-8 -*-
#
# Copyright 2019 Francesco Magno
# Copyright 2019 Kenneth Loafman
#
# This file is part of duplicity.
#
# Duplicity is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# Duplicity is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with duplicity; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

"""
Rclone backend.
"""

import os
import os.path

import duplicity.backend
from duplicity import errors
from duplicity import log
from duplicity.errors import BackendException


class RcloneBackend(duplicity.backend.Backend):
    """
    Backend that uses rclone to access remote storage.
    """

    def __init__(self, parsed_url):
        """
        Initialize the rclone backend.

        :param ParsedUrl parsed_url: The parsed URL.
        :raises FatalError: If rclone is not found.
        """
        duplicity.backend.Backend.__init__(self, parsed_url)
        self.parsed_url = parsed_url
        self.remote_path = self.parsed_url.path
        self.rclone_cmd = "rclone"

        try:
            rc, o, e = self._subprocess_safe_popen(f"{self.rclone_cmd} version")
        except Exception:
            raise errors.FatalError("rclone not found: please install rclone", log.ErrorCode.backend_error)

        verb = log.getverbosity()
        if verb >= log.DEBUG:
            os.environ["RCLONE_LOG_LEVEL"] = "DEBUG"
        elif verb >= log.INFO:
            os.environ["RCLONE_LOG_LEVEL"] = "INFO"
        elif verb >= log.NOTICE:
            os.environ["RCLONE_LOG_LEVEL"] = "NOTICE"
        elif verb >= log.ERROR:
            os.environ["RCLONE_LOG_LEVEL"] = "ERROR"

        if parsed_url.path.startswith("//"):
            self.remote_path = self.remote_path[2:].replace(":/", ":", 1)

        self.remote_path = os.fsdecode(self.remote_path)

    def _get(self, remote_filename, local_path):
        remote_filename = os.fsdecode(remote_filename)
        local_pathname = os.fsdecode(local_path.name)
        commandline = f"{self.rclone_cmd} copyto '{self.remote_path}/{remote_filename}' '{local_pathname}'"
        rc, o, e = self._subprocess_safe_popen(commandline)
        if rc != 0:
            if os.path.isfile(local_pathname):
                os.remove(local_pathname)
            raise BackendException(f"rclone returned rc = {int(rc)}: {e}")

    def _put(self, source_path, remote_filename):
        source_pathname = os.fsdecode(source_path.name)
        remote_filename = os.fsdecode(remote_filename)
        commandline = f"{self.rclone_cmd} copyto '{source_pathname}' '{self.remote_path}/{remote_filename}'"
        rc, o, e = self._subprocess_safe_popen(commandline)
        if rc != 0:
            raise BackendException(f"rclone returned rc = {int(rc)}: {e}")

    def _list(self):
        """
        List files in the remote rclone path.

        :returns: A list of filenames as byte strings.
        :rtype: list
        :raises BackendException: If rclone returns an error.
        """
        filelist = []
        commandline = f"{self.rclone_cmd} lsf '{self.remote_path}'"
        rc, o, e = self._subprocess_safe_popen(commandline)
        if rc == 3:
            return filelist
        if rc != 0:
            raise BackendException(f"rclone returned rc = {int(rc)}: {e}")
        if not o:
            return filelist
        return [os.fsencode(x) for x in o.split("\n") if x]

    def _query(self, remote_filename: str):
        """
        Get metadata for a file using rclone.

        :param str remote_filename: The name of the file to query.
        :returns: A dictionary containing metadata (e.g., 'size').
        :rtype: dict
        """
        size = None
        try:
            remote_filename = os.fsdecode(remote_filename)
            commandline = f"{self.rclone_cmd} lsf --format=s '{self.remote_path}/{remote_filename}'"
            rc, o, e = self._subprocess_safe_popen(commandline)
            if rc == 3:  # not found
                size = -1
            elif rc == 0:
                size = int(o)
        except Exception as e:
            log.Debug(f"Failed to query file size for {remote_filename}: {e}")
            return {"size": None}
        return {"size": size}

    def _delete(self, remote_filename):
        """
        Delete a file from the remote rclone path.

        :param str remote_filename: The name of the file to delete.
        :raises BackendException: If rclone returns an error.
        """
        remote_filename = os.fsdecode(remote_filename)
        commandline = f"{self.rclone_cmd} deletefile --drive-use-trash=false '{self.remote_path}/{remote_filename}'"
        rc, o, e = self._subprocess_safe_popen(commandline)
        if rc != 0:
            raise BackendException(f"rclone returned rc = {int(rc)}: {e}")

    def _subprocess_safe_popen(self, commandline):
        import shlex
        from subprocess import Popen
        from subprocess import PIPE

        args = shlex.split(commandline)
        p = Popen(args, stdout=PIPE, stderr=PIPE, universal_newlines=True)
        stdout, stderr = p.communicate()
        for l in stderr.split("\n"):
            if len(l) > 1:
                print(l)
        return p.returncode, stdout, stderr


duplicity.backend.register_backend("rclone", RcloneBackend)



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4; coding:utf-8 -*-
#
# Copyright 2022 Kenneth Loafman
#
# This file is part of duplicity.
#
# Duplicity is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# Duplicity is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with duplicity; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

"""
Utils for parse command line, check for consistency, and set config
"""

import io
import os
import re
import socket
import sys
from hashlib import md5
from textwrap import dedent

# TODO: Remove duplicity.argparse311 when py310 goes EOL
from duplicity import argparse311 as argparse
from duplicity import config
from duplicity import dup_time
from duplicity import errors
from duplicity import log
from duplicity import path
from duplicity import selection

# gpg_key_patt = re.compile(r"^(0x)?([0-9A-Fa-f]{8}|[0-9A-Fa-f]{16}|[0-9A-Fa-f]{40})$")
url_regexp = re.compile(r"^[\w\+]+://")

help_footer = _("Enter 'duplicity --help' for help screen.")


class CommandLineError(errors.UserError):
    """
    Exception raised when a command line error occurs.
    """

    # sys.tracebacklimit = 4
    pass


def command_line_error(message):
    """
    Indicate a command line error and exit.

    :param str message: The error message.
    :raises CommandLineError: Always.
    """
    raise CommandLineError(f"{message}\n{help_footer}")


class DuplicityAction(argparse.Action):
    """
    Base class for custom argparse actions in Duplicity.
    """

    def __init__(self, option_strings, dest, **kwargs):
        super().__init__(option_strings, dest, **kwargs)

    def __call__(self, parser, namespace, values, option_string=None):
        raise NotImplementedError


class DoNothingAction(DuplicityAction):
    """
    Argparse action that does nothing.
    """

    def __call__(self, parser, *args, **kw):
        pass


class AddSelectionAction(DuplicityAction):
    """
    Argparse action that adds a selection option to the configuration.
    """

    def __init__(self, option_strings, dest, **kwargs):
        super().__init__(option_strings, dest, **kwargs)

    def __call__(self, parser, namespace, values, option_string=None):
        addarg = os.fsdecode(values) if isinstance(values, bytes) else values
        if addarg == "":
            command_line_error(f"Option {option_string} cannot be empty.")
        config.select_opts.append((os.fsdecode(option_string), addarg))


class AddFilelistAction(DuplicityAction):
    """
    Argparse action that adds a filelist to the configuration.
    """

    def __init__(self, option_strings, dest, **kwargs):
        super().__init__(option_strings, dest, **kwargs)

    def __call__(self, parser, namespace, values, option_string=None):
        config.select_opts.append((os.fsdecode(option_string), os.fsdecode(values)))
        try:
            config.select_files.append(io.open(values, "rt", encoding="UTF-8"))
        except Exception as e:
            command_line_error(str(e))


class AddRenameAction(DuplicityAction):
    """
    Argparse action that adds a rename rule to the configuration.
    """

    def __init__(self, option_strings, dest, **kwargs):
        super().__init__(option_strings, dest, **kwargs)

    def __call__(self, parser, namespace, values, option_string=None):
        key = os.fsencode(os.path.normcase(os.path.normpath(values[0])))
        config.rename[key] = os.fsencode(values[1])


class SplitOptionsAction(DuplicityAction):
    """
    Argparse action that splits and appends options to a string.
    """

    def __init__(self, option_strings, dest, **kwargs):
        super().__init__(option_strings, dest, **kwargs)

    def __call__(self, parser, namespace, values, option_string=None):
        var = opt2var(option_string)
        opts = getattr(namespace, var)
        values = values.strip('"').strip("'")
        if opts == "":
            opts = values
        else:
            opts = f"{opts} {values}"
        setattr(namespace, var, opts)


class IgnoreErrorsAction(DuplicityAction):
    """
    Argparse action that enables error ignoring and logs a warning.
    """

    def __init__(self, option_strings, dest, **kwargs):
        super().__init__(option_strings, dest, **kwargs)

    def __call__(self, parser, namespace, values, option_string=None):
        var = opt2var(option_string)
        log.Warn(
            _("Running in 'ignore errors' mode due to --ignore-errors.\n" "Please reconsider if this was not intended")
        )
        setattr(namespace, var, True)


class WarnAsyncStoreConstAction(argparse._StoreConstAction):
    """
    Argparse action that warns about deprecated asynchronous-upload option.
    """

    def __init__(self, option_strings, dest, nargs=None, **kwargs):
        super().__init__(option_strings, dest, **kwargs)

    def __call__(self, parser, namespace, values, option_string=None):
        log.Warn(_(dedent("""
                    ----------------------------------------------------------------
                    | WARNING: replaced with `--concurrency`                       |
                    | Use of the --asynchronous-upload option was known to be      |
                    | unsafe and may result in data loss.                          |
                    | It was removed in VERSION 3.0.0 and replaced by              |
                    | --concurrency which will offer similar functionality, but    |
                    | thoroughly tested implementation.                            |
                    | See: https://gitlab.com/duplicity/duplicity/-/issues/745 and |
                    | https://gitlab.com/duplicity/duplicity/-/merge_requests/153  |
                    ----------------------------------------------------------------
                    """)))
        setattr(namespace, self.dest, self.const)


class SetLogTimestampAction(argparse._StoreConstAction):
    def __init__(self, option_strings, dest, nargs=None, **kwargs):
        super().__init__(option_strings, dest, **kwargs)

    def __call__(self, parser, namespace, values, option_string=None):
        log.add_timestamp()


def _check_int(val):
    try:
        return int(val)
    except Exception as e:
        command_line_error(_(f"'{val}' is not an int: {str(e)}"))


def _check_time(val):
    try:
        return dup_time.genstrtotime(val)
    except dup_time.TimeException as e:
        command_line_error(str(e))


def check_char(val):
    if len(val) == 1:
        return val
    else:
        command_line_error(_(f"'{val} is not a single character."))


def check_count(val):
    return _check_int(val)


def check_file(val):
    try:
        return os.fsencode(expand_fn(val))
    except Exception as e:
        command_line_error(f"{val} is not a valid pathname: {str(e)}")


def check_time_delta(val):
    try:
        return dup_time.genstrtoseconds(val)
    except dup_time.TimeException as e:
        command_line_error(str(e))


def check_remove_time(val):
    return _check_time(val)


def check_source_path(val):
    if not is_path(val):
        command_line_error(_(f"Source should be pathname, not url.  Got '{val}' instead."))
    if not os.path.exists(val):
        command_line_error(_(f"Argument source_path '{val}' does not exist."))
    return val


def check_source_url(val):
    if not is_url(val):
        command_line_error(_(f"Source should be url, not directory.  Got '{val}' instead."))
    return val


def check_target_path(val):
    if not is_path(val):
        command_line_error(_(f"Target should be directory, not url.  Got '{val}' instead."))
    if not os.path.exists(val):
        try:
            os.makedirs(val, exist_ok=True)
        except Exception as e:
            command_line_error(_(f"Unable to create target dir '{val}': {str(e)}"))
    return val


def check_target_url(val):
    if not is_url(val):
        command_line_error(_(f"Target should be url, not directory.  Got '{val}' instead."))
    return val


def check_time(val):
    return _check_time(val)


def check_timeout(val):
    """
    set timeout for backends
    """
    val = _check_int(val)
    socket.setdefaulttimeout(val)
    return val


def check_verbosity(val):
    fail = False
    verb = log.NOTICE
    val = val.lower()
    if val in ["e", "error"]:
        verb = log.ERROR
    elif val in ["w", "warning"]:
        verb = log.WARNING
    elif val in ["n", "notice"]:
        verb = log.NOTICE
    elif val in ["i", "info"]:
        verb = log.INFO
    elif val in ["d", "debug"]:
        verb = log.DEBUG
    else:
        try:
            verb = int(val)
            if verb < 0 or verb > 9:
                fail = True
        except ValueError:
            fail = True

    if fail:
        # TRANSL: In this portion of the usage instructions, "[ewnid]" indicates which
        # characters are permitted (e, w, n, i, or d); the brackets imply their own
        # meaning in regex; i.e., only one of the characters is allowed in an instance.
        command_line_error(
            _(
                "Verbosity must be one of: digit [0-9], character [ewnid],\n"
                "or word ['error', 'warning', 'notice', 'info', 'debug'].\n"
                "The default is 3 (Notice).  It is strongly recommended\n"
                "that verbosity level is set at 2 (Warning) or higher."
            )
        )

    log.setverbosity(verb)
    config.verbosity = verb
    return verb


def dflt(val):
    """
    Return printable value for default.
    """
    return val


def expand_fn(filename):
    """
    Expand user and vars in filename
    """
    return os.path.expanduser(os.path.expandvars(filename))


def expand_archive_dir(archdir, backname):
    """
    Return expanded version of archdir joined with backname.
    """
    assert config.backup_name is not False, "expand_archive_dir() called prior to config.backup_name being set"

    return expand_fn(os.path.join(archdir, os.fsencode(backname)))


def generate_default_backup_name(backend_url):
    """
    Generate a default backup name based on the backend URL.

    :param str backend_url: The URL to the backend.
    :returns: A default backup name as a hexadecimal hash.
    :rtype: str
    """
    # For default, we hash args to obtain a reasonably safe default.
    # We could be smarter and resolve things like relative paths, but
    # this should actually be a pretty good compromise. Normally only
    # the destination will matter since you typically only restart
    # backups of the same thing to a given destination. The inclusion
    # of the source however, does protect against most changes of
    # source directory (for whatever reason, such as
    # /path/to/different/snapshot). If the user happens to have a case
    # where relative paths are used yet the relative path is the same
    # (but duplicity is run from a different directory or similar),
    # then it is simply up to the user to set --archive-dir properly.
    burlhash = md5(usedforsecurity=False)
    burlhash.update(backend_url.encode())
    return burlhash.hexdigest()


def is_url(val):
    """
    Check if the given value is a URL.

    :param str val: The value to check.
    :returns: True if the value looks like a URL, False otherwise.
    :rtype: bool
    """
    return len(val.splitlines()) <= 1 and url_regexp.match(val)


def is_path(val):
    """
    Check if the given value is a file path.

    :param str val: The value to check.
    :returns: True if the value looks like a path, False otherwise.
    :rtype: bool
    """
    return not is_url(val)


def make_bytes(value):
    if isinstance(value, str):
        return bytes(value, "utf-8")


def var2cmd(s):
    """
    Convert a variable name to a command string (replaces underscores with hyphens).

    :param str s: The variable name.
    :returns: The command string.
    :rtype: str
    """
    return s.replace("_", "-")


def var2opt(s):
    """
    Convert a variable name to an option string.

    :param str s: The variable name.
    :returns: The option string (prefixed with - or --).
    :rtype: str
    """
    if len(s) > 1:
        return f"--{s.replace('_', '-')}"
    else:
        return f"-{s}"


def cmd2var(s):
    """
    Convert command string to var name
    """
    return s.replace("-", "_")


def opt2var(s):
    """
    Convert option string to var name
    """
    return s.lstrip("-").replace("-", "_")


def set_log_fd(fd):
    try:
        fd = int(fd)
    except ValueError:
        command_line_error("log_fd must be an integer.")
    if fd < 1:
        command_line_error("log-fd must be greater than zero.")
    log.add_fd(fd)
    return fd


def set_log_file(fn):
    fn = check_file(fn)
    log.add_file(fn)
    return fn


def set_kilos(num):
    return _check_int(num) * 1024


def set_megs(num):
    return _check_int(num) * 1024 * 1024


def set_archive_dir(dirstring):
    """
    Check archive dir and set global
    """
    if not os.path.exists(dirstring):
        try:
            os.makedirs(dirstring)
        except Exception:
            pass
    archive_dir_path = path.Path(dirstring)
    if not archive_dir_path.isdir():
        command_line_error(_(f"Specified archive directory '{archive_dir_path.uc_name}' is not a directory"))
    config.archive_dir_path = archive_dir_path


def set_encrypt_key(encrypt_key):
    """
    Set config.gpg_profile.encrypt_key assuming proper key given
    """
    if config.gpg_profile.recipients is None:
        config.gpg_profile.recipients = []
    config.gpg_profile.recipients.append(encrypt_key)


def set_encrypt_sign_key(encrypt_sign_key):
    """
    Set config.gpg_profile.encrypt_sign_key assuming proper key given
    """
    set_encrypt_key(encrypt_sign_key)
    set_sign_key(encrypt_sign_key)


def set_hidden_encrypt_key(hidden_encrypt_key):
    """
    Set config.gpg_profile.hidden_encrypt_key assuming proper key given
    """
    if config.gpg_profile.hidden_recipients is None:
        config.gpg_profile.hidden_recipients = []
    config.gpg_profile.hidden_recipients.append(hidden_encrypt_key)


def set_sign_key(sign_key):
    """
    Set config.gpg_profile.sign_key assuming proper key given
    """
    config.gpg_profile.sign_key = sign_key


def set_selection():
    """
    Return selection iter starting at filename with arguments applied
    """
    sel = selection.Select(config.local_path)
    sel.ParseArgs(config.select_opts, config.select_files)
    config.select = sel.set_iter()


def round512(s):
    """
    Rounds up to the next 512 boundary. For negative numbers the absolute value will be used.
    Minimum return value is 512.
    """
    return ((max(1, abs(int(s))) + 511) // 512) * 512



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4; coding:utf-8 -*-
#
# Copyright 2002 Ben Escoto
# Copyright 2007 Kenneth Loafman
#
# This file is part of duplicity.
#
# Duplicity is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# Duplicity is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with duplicity; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

"""
Rsync backend.
"""

import os
import re
import tempfile

import duplicity.backend
from duplicity import config
from duplicity import tempdir
from duplicity import util
from duplicity.errors import InvalidBackendURL


class RsyncBackend(duplicity.backend.Backend):
    """Connect to remote store using rsync

    rsync backend contributed by Sebastian Wilhelmi <seppi@seppi.de>
    rsyncd auth, alternate port support
    Copyright 2010 by Edgar Soldin
    """

    def __init__(self, parsed_url):
        """
        Initialize the rsync backend.

        :param ParsedUrl parsed_url: The parsed URL.
        :raises InvalidBackendURL: If the URL is invalid.
        """
        duplicity.backend.Backend.__init__(self, parsed_url)
        """
        rsyncd module url: rsync://[user:password@]host[:port]::[/]modname/path
                      Note: 3.0.7 is picky about syntax use either 'rsync://' or '::'
                      cmd: rsync [--port=port] host::modname/path
        -or-
        rsync via ssh/rsh url: rsync://user@host[:port]://some_absolute_path
             -or-              rsync://user@host[:port]:/some_relative_path
                          cmd: rsync -e 'ssh [-p=port]' [user@]host:[/]path
        """
        host = parsed_url.hostname
        port = ""
        # RSYNC_RSH from calling shell might conflict with our settings
        if "RSYNC_RSH" in os.environ:
            del os.environ["RSYNC_RSH"]
        if self.over_rsyncd():
            # its a module path
            path, port = self.get_rsync_path()
            self.url_string = f"{host}::{path.lstrip('/:')}"
            if port:
                port = f" --port={port}"
        else:
            host_string = f"{host}:" if host else ""
            if parsed_url.path.startswith("//"):
                # its an absolute path
                self.url_string = f"{host_string}/{parsed_url.path.lstrip('/')}"
            else:
                # its a relative path
                self.url_string = f"{host_string}{parsed_url.path.lstrip('/')}"
            if parsed_url.port:
                port = f"-p {parsed_url.port}"
        # add trailing slash if missing
        if self.url_string[-1] != "/":
            self.url_string += "/"
        # user?
        if parsed_url.username:
            if self.over_rsyncd():
                os.environ["USER"] = parsed_url.username
            else:
                self.url_string = f"{parsed_url.username}@{self.url_string}"
        # password?, don't ask if none was given
        self.use_getpass = False
        password = self.get_password()
        if password:
            os.environ["RSYNC_PASSWORD"] = password
        if self.over_rsyncd():
            portOption = port
        else:
            portOption = f"-e 'ssh {port} -oBatchMode=yes {config.ssh_options}'"
        rsyncOptions = config.rsync_options
        # build cmd
        self.cmd = f"rsync {portOption} {rsyncOptions}"

    def over_rsyncd(self):
        url = self.parsed_url.url_string
        if re.search("::[^:]*$", url):
            return True
        else:
            return False

    def get_rsync_path(self):
        url = self.parsed_url.url_string
        m = re.search(r"(:\d+|)?::([^:]*)$", url)
        if m:
            return m.group(2), m.group(1).lstrip(":")
        raise InvalidBackendURL(f"Could not determine rsync path: {self.munge_password(url)}")

    def _put(self, source_path, remote_filename):
        remote_filename = os.fsdecode(remote_filename)
        remote_path = os.path.join(self.url_string, remote_filename)
        commandline = f"{self.cmd} {source_path.uc_name} {remote_path}"
        self.subprocess_popen(commandline)

    def _get(self, remote_filename, local_path):
        remote_filename = os.fsdecode(remote_filename)
        remote_path = os.path.join(self.url_string, remote_filename)
        commandline = f"{self.cmd} {remote_path} {local_path.uc_name}"
        self.subprocess_popen(commandline)

    def _list(self):
        """
        List files in the remote rsync path.

        :returns: A list of filenames as byte strings.
        :rtype: list
        """

        def split(str):  # pylint: disable=redefined-builtin
            line = str.split()
            if len(line) > 4 and line[4] != ".":
                return line[4]
            else:
                return None

        commandline = f"{self.cmd} {self.url_string}"
        result, stdout, stderr = self.subprocess_popen(commandline)
        return [os.fsencode(x) for x in map(split, stdout.split("\n")) if x]

    def _delete_list(self, filename_list):
        """
        Delete a list of files from the remote rsync path.

        :param list filename_list: The names of the files to delete.
        """
        delete_list = filename_list
        dont_delete_list = []
        for file in self._list():
            if file in delete_list:
                delete_list.remove(file)
            else:
                dont_delete_list.append(file)

        dir = tempfile.mkdtemp()  # pylint: disable=redefined-builtin
        exclude, exclude_name = tempdir.default().mkstemp_file()
        to_delete = [exclude_name]
        for file in dont_delete_list:
            file = os.fsdecode(file)
            path = os.path.join(dir, file)
            to_delete.append(path)
            try:
                f = open(path, "w")
            except IsADirectoryError:
                print(file, file=exclude)
                continue
            print(file, file=exclude)
            f.close()
        exclude.close()
        commandline = f"{self.cmd} --recursive --delete --exclude-from={exclude_name} {dir}/ {self.url_string}"
        self.subprocess_popen(commandline)
        for file in to_delete:
            try:
                util.ignore_missing(os.unlink, file)
            except IsADirectoryError:
                pass
        os.rmdir(dir)


duplicity.backend.register_backend("rsync", RsyncBackend)
duplicity.backend.uses_netloc.extend(["rsync"])



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4; coding:utf-8 -*-
#
# Copyright 2002 Ben Escoto
# Copyright 2007 Kenneth Loafman
#
# This file is part of duplicity.
#
# Duplicity is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# Duplicity is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with duplicity; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

"""
HSI (HPSS) backend.
"""

import os

import duplicity.backend

hsi_command = "hsi"


class HSIBackend(duplicity.backend.Backend):
    def __init__(self, parsed_url):
        """
        Initialize the HSI backend.

        :param ParsedUrl parsed_url: The parsed URL.
        """
        duplicity.backend.Backend.__init__(self, parsed_url)
        self.host_string = parsed_url.hostname
        self.remote_dir = parsed_url.path
        if self.remote_dir:
            self.remote_prefix = f"{self.remote_dir}/"
        else:
            self.remote_prefix = ""

    def _put(self, source_path, remote_filename):
        if isinstance(remote_filename, bytes):
            remote_filename = os.fsdecode(remote_filename)
        commandline = f'{hsi_command} "put {source_path.uc_name} : {self.remote_prefix}{remote_filename}"'
        self.subprocess_popen(commandline)

    def _get(self, remote_filename, local_path):
        if isinstance(remote_filename, bytes):
            remote_filename = os.fsdecode(remote_filename)
        commandline = f'{hsi_command} "get {local_path.uc_name} : {self.remote_prefix}{remote_filename}"'
        self.subprocess_popen(commandline)

    def _list(self):
        """
        List files in the remote HSI folder.

        :returns: A list of filenames as byte strings.
        :rtype: list
        """
        commandline = f'{hsi_command} "ls -l {self.remote_dir}"'
        l = self.subprocess_popen(commandline)[2]
        l = l.split(os.linesep.encode())[3:]
        for i in range(0, len(l)):
            if l[i]:
                l[i] = l[i].split()[-1]
        return [os.fsencode(x) for x in l if x]

    def _delete(self, filename):
        if isinstance(filename, bytes):
            filename = os.fsdecode(filename)
        commandline = f'{hsi_command} "rm {self.remote_prefix}{filename}"'
        self.subprocess_popen(commandline)


duplicity.backend.register_backend("hsi", HSIBackend)
duplicity.backend.uses_netloc.extend(["hsi"])



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4; coding:utf-8 -*-
#
# Copyright 2013 J.P. Krauss
#
# This file is part of duplicity.
#
# Duplicity is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# Duplicity is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with duplicity; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

"""
Rackspace CloudFiles backend using Pyrax.
"""

import os

import duplicity.backend
from duplicity import errors
from duplicity import log
from duplicity import util
from duplicity.errors import BackendException


class PyraxBackend(duplicity.backend.Backend):
    """
    Backend for Rackspace's CloudFiles using Pyrax
    """

    def __init__(self, parsed_url):
        """
        Initialize the Pyrax-based CloudFiles backend.

        :param ParsedUrl parsed_url: The parsed URL.
        """
        duplicity.backend.Backend.__init__(self, parsed_url)

        try:
            import pyrax
        except ImportError as e:
            raise BackendException(f"""Pyrax backend requires the pyrax library available from Rackspace.
Exception: {str(e)}""")

        # Inform Pyrax that we're talking to Rackspace
        # per Jesus Monzon (gsusmonzon)
        pyrax.set_setting("identity_type", "rackspace")

        conn_kwargs = {}

        if "CLOUDFILES_USERNAME" not in os.environ:
            raise BackendException("CLOUDFILES_USERNAME environment variable" "not set.")

        if "CLOUDFILES_APIKEY" not in os.environ:
            raise BackendException("CLOUDFILES_APIKEY environment variable not set.")

        conn_kwargs["username"] = os.environ["CLOUDFILES_USERNAME"]
        conn_kwargs["api_key"] = os.environ["CLOUDFILES_APIKEY"]

        if "CLOUDFILES_REGION" in os.environ:
            conn_kwargs["region"] = os.environ["CLOUDFILES_REGION"]

        container = parsed_url.path.lstrip("/")

        try:
            pyrax.set_credentials(**conn_kwargs)
        except Exception as e:
            raise errors.FatalError(
                f"Connection failed, please check your credentials: {e.__class__.__name__} {util.uexc(e)}",
                log.ErrorCode.connection_failed,
            )

        self.client_exc = pyrax.exceptions.ClientException
        self.nso_exc = pyrax.exceptions.NoSuchObject

        # query rackspace for the specified container name
        try:
            self.container = pyrax.cloudfiles.get_container(container)
        except pyrax.exceptions.Forbidden as e:
            raise errors.FatalError(
                f"{e.__class__.__name__} : {util.uexc(e)} \n"
                + "Container may exist, but access was denied.\n"
                + "If this container exists, please check its X-Container-Read/Write headers.\n"
                + "Otherwise, please check your credentials and permissions.",
                log.ErrorCode.backend_permission_denied,
            )
        except pyrax.exceptions.NoSuchContainer as e:
            try:
                self.container = pyrax.cloudfiles.create_container(container)
            except pyrax.exceptions.Forbidden as e:
                raise errors.FatalError(
                    f"{e.__class__.__name__} : {util.uexc(e)} \n"
                    + "Container does not exist, but creation was denied.\n"
                    + "You may be using a read-only user that can view but not create containers.\n"
                    + "Please check your credentials and permissions.",
                    log.ErrorCode.backend_permission_denied,
                )

    def _error_code(self, operation, e):  # pylint: disable=unused-argument
        if isinstance(e, self.nso_exc):
            return log.ErrorCode.backend_not_found
        elif isinstance(e, self.client_exc):
            if e.code == 404:
                return log.ErrorCode.backend_not_found
        elif hasattr(e, "http_status"):
            if e.http_status == 404:
                return log.ErrorCode.backend_not_found

    def _put(self, source_path, remote_filename):
        self.container.upload_file(source_path.name, remote_filename)

    def _get(self, remote_filename, local_path):
        sobject = self.container.get_object(remote_filename)
        with open(local_path.name, "wb") as f:
            f.write(sobject.get())

    def _list(self):
        """
        List files in the CloudFiles container using Pyrax.

        Handles the 10,000 object limit by making multiple requests.

        :returns: A list of filenames.
        :rtype: list
        """
        # Cloud Files will return a max of 10,000 objects.  We have
        # to make multiple requests to get them all.
        objs = self.container.get_object_names()
        keys = objs
        while len(objs) == 10000:
            objs = self.container.get_object_names(marker=keys[-1])
            keys += objs
        return keys

    def _delete(self, filename):
        self.container.delete_object(filename)

    def _query(self, filename):
        sobject = self.container.get_object(filename)
        return {"size": sobject.total_bytes}



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4; coding:utf-8 -*-
#
# Copyright 2002 Ben Escoto
# Copyright 2007 Kenneth Loafman
#
# This file is part of duplicity.
#
# Duplicity is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# Duplicity is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with duplicity; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

"""
Test backend that can simulate various failures.
"""

import glob
import hashlib
import inspect
import json
import logging
import os
import re
import sys
import time
from random import random

import duplicity.backend
from duplicity import errors
from duplicity import log
from duplicity import path
from duplicity import progress
from duplicity.errors import BackendException


class BackendErrors:
    # FAIL_WITH_EXCEPTION: set to substring matched by the filename.
    FAIL_WITH_EXCEPTION = "DUP_FAIL_WITH_EXCEPTION"
    # SYSTEM_EXIT: set to substring matched by the filename.
    FAIL_SYSTEM_EXIT = "DUP_FAIL_BY_SYSTEM_EXIT"
    # WAIT_FOR_OTHER_VOLUME: set to json string: ["file_to_delay", "file_wait_for"] (substring match)
    WAIT_FOR_OTHER_VOLUME = "DUP_FAIL_WAIT_FOR_VOLUME"
    # LAST_BYTE_MISSING: set to substring matched by the filename.
    LAST_BYTE_MISSING = "DUP_FAIL_LAST_BYTE_MISSING"
    # SKIP_PUT_SILENT: Don't put file if name contains string
    SKIP_PUT_SILENT = "DUP_FAIL_SKIP_PUT_SILENT"
    # RANDOM_DELAY: max random delay in ms added to command
    DELAY_RANDOM_MS = "DUP_FAIL_DELAY_RANDOM_MS"
    # FIX_DELAY: fix delay in ms added to command
    DELAY_FIX_MS = "DUP_FAIL_DELAY_FIX_MS"


class _TestBackend(duplicity.backend.Backend):
    """Use this backend to test/create certain error situations

    errors get triggered via ENV Vars, see class BackendErrors

    """

    def __init__(self, parsed_url):
        """
        Initialize the test backend.

        :param ParsedUrl parsed_url: The parsed URL.
        :raises BackendException: If the URL path syntax is invalid.
        """
        super().__init__(parsed_url)
        from testing import _runtest_dir

        log._logger.addHandler(logging.FileHandler(f"{_runtest_dir}/testbackend.log"))
        log.Warn("TestBackend is not made for production use!")
        # The URL form "file:MyFile" is not a valid duplicity target.
        if not parsed_url.path.startswith("//"):
            raise BackendException("Bad file:// path syntax.")
        self.remote_pathdir = path.Path(parsed_url.path[2:])
        try:
            os.makedirs(self.remote_pathdir.base)
        except Exception:
            pass

    @staticmethod
    def _fail_with_exception(filename):
        filename = os.fsdecode(filename)
        log.Debug(f"DUB_FAIL: Check if fail on exception for {filename}. Called by {inspect.stack()[2][3]}.")
        if os.getenv(BackendErrors.FAIL_WITH_EXCEPTION) and os.getenv(BackendErrors.FAIL_WITH_EXCEPTION) in filename:
            log.Error(f"DUB_FAIL: Force exception on file {filename}.")
            raise FileNotFoundError(f"TEST: raised exception on {filename} by intention")

    @staticmethod
    def _fail_by_sys_exit(filename):
        filename = os.fsdecode(filename)
        log.Debug(f"DUB_FAIL: Check if do sys.exit(30) for {filename}. Called by {inspect.stack()[2][3]}.")
        if os.getenv(BackendErrors.FAIL_SYSTEM_EXIT) and os.getenv(BackendErrors.FAIL_SYSTEM_EXIT) in filename:
            log.Error(f"DUB_FAIL: Force sys.exit(30) on file {filename}.")
            time.sleep(1)  # otherwise log message doesn't get printed.
            sys.exit(30)

    def _wait_for_other_volume(self, filename):
        filename = os.fsdecode(filename)
        log.Debug(f"DUB_FAIL: Check if action on {filename} shoud be delayed. Called by {inspect.stack()[2][3]}.")
        env = os.getenv(BackendErrors.WAIT_FOR_OTHER_VOLUME)
        if not env:
            return  # skip if env not set
        file_stop, file_waitfor = json.loads(env)
        timestamp = re.match(r".*\.(\d{8}T\d{6}[-+0-9:Z]*)\..*", filename).group(1)
        file_waitfor_glob = f"{os.fsdecode(self.remote_pathdir.get_canonical())}/*{timestamp}*{file_waitfor}*"
        if file_stop in filename:
            while not glob.glob(file_waitfor_glob):
                log.Error(f"DUB_FAIL: Waiting for file matiching {file_waitfor_glob}.")
                time.sleep(1)
            log.Warn(f"DUB_FAIL: {filename} written after {glob.glob(file_waitfor_glob)}")

    def _remove_last_byte(self, filename):
        filename = os.fsdecode(filename)
        log.Debug(f"DUB_FAIL: Check if {filename} shoud be truncated. Called by {inspect.stack()[2][3]}.")
        if os.getenv(BackendErrors.LAST_BYTE_MISSING) and os.getenv(BackendErrors.LAST_BYTE_MISSING) in filename:
            log.Error(f"DUB_FAIL: removing last byte from {filename}")
            with open(self.remote_pathdir.append(filename).get_canonical(), "ab") as remote_file:
                remote_file.seek(-1, os.SEEK_END)
                remote_file.truncate()

    @staticmethod
    def _skip_put_silent(filename):
        """
        retrun true if file should be skipped silently
        """
        filename = os.fsdecode(filename)
        log.Debug(f"DUB_FAIL: Check if {filename} should be skipped. Called by {inspect.stack()[2][3]}.")
        if os.getenv(BackendErrors.SKIP_PUT_SILENT) and os.getenv(BackendErrors.SKIP_PUT_SILENT) in filename:
            log.Error(f"DUB_FAIL: {filename} skipped silent.")
            return True
        return False

    @staticmethod
    def _delay():
        """
        sleep a random amount of milliseconds if ENV set
        """
        log.Debug("DUB_FAIL: Check if action should be delayed.")
        wait = 0
        if os.getenv(BackendErrors.DELAY_RANDOM_MS):
            wait += random() * float(os.getenv(BackendErrors.DELAY_RANDOM_MS)) / 1000  # type: ignore
        if os.getenv(BackendErrors.DELAY_FIX_MS):
            wait += float(os.getenv(BackendErrors.DELAY_FIX_MS)) / 1000  # type: ignore
        if wait > 0:
            log.Warn(f"DUB_FAIL: wait for {wait} sec.")
            time.sleep(wait)

    def _move(self, source_path, remote_filename):
        self._fail_with_exception(remote_filename)
        self._fail_by_sys_exit(remote_filename)
        self._wait_for_other_volume(remote_filename)
        self._delay()
        target_path = self.remote_pathdir.append(remote_filename)
        try:
            source_path.rename(target_path)
            return True
        except OSError:
            return False

    def _put(self, source_path, remote_filename):
        self._fail_with_exception(remote_filename)
        self._wait_for_other_volume(remote_filename)
        self._delay()
        if self._skip_put_silent(remote_filename):
            return
        target_path = self.remote_pathdir.append(remote_filename)
        source_path.setdata()
        source_size = source_path.getsize()
        progress.report_transfer(0, source_size)
        target_path.writefileobj(source_path.open("rb"))

        self._fail_by_sys_exit(remote_filename)  # fail after file is transferred
        self._remove_last_byte(remote_filename)

        progress.report_transfer(source_size, source_size)

    def _get(self, filename, local_path):
        self._fail_with_exception(filename)
        self._fail_by_sys_exit(filename)
        self._delay()
        source_path = self.remote_pathdir.append(filename)
        local_path.writefileobj(source_path.open("rb"))

    def _list(self):
        return self.remote_pathdir.listdir()

    def _validate(self, remote_filename, size, source_path=None, **kwargs):
        # poc to show that validate can do additional things.
        self._delay()

        self._remove_last_byte(remote_filename)

        results_str = []
        results_bool = []
        try:
            if source_path:
                target_path = self.remote_pathdir.append(remote_filename)
                target_hash = self.__hash_fileobj(target_path.open())
                source_hash = self.__hash_fileobj(source_path.open())
                if target_hash == source_hash:
                    results_str.append(f"file hash {target_hash} matches")
                    results_bool.append(True)
                else:
                    results_str.append(f"expected hash {source_hash} doesn't match file hash {target_hash}")
                    results_bool.append(False)
            if size == self._query(remote_filename)["size"]:
                results_str.append(f"file size {size}")
                results_bool.append(True)
            else:
                results_str.append(
                    f'expected size {size} and file size {self._query(remote_filename)["size"]} don\'t match'
                )
                results_bool.append(False)
        except FileNotFoundError as e:
            results_bool.append(False)
            results_str.append(f"FileNotFoundError: {e}")
        except Exception as e:
            raise errors.FatalError(
                _("Unexpected exception while validate %s.") % os.fsdecode(remote_filename),
                log.ErrorCode.backend_validation_failed,
            )
        return all(results_bool), ", ".join(results_str)

    def _delete(self, filename):
        self._fail_with_exception(filename)
        self._fail_by_sys_exit(filename)
        self._delay()
        self.remote_pathdir.append(filename).delete()

    def _delete_list(self, filenames):
        for filename in filenames:
            self._fail_with_exception(filename)
            self._fail_by_sys_exit(filename)
            self._delay()
            self.remote_pathdir.append(filename).delete()

    def _query(self, filename):
        self._fail_with_exception(filename)
        self._delay()
        self._fail_by_sys_exit(filename)
        target_file = self.remote_pathdir.append(filename)
        target_file.setdata()
        size = target_file.getsize() if target_file.exists() else -1
        return {"size": size}

    def _query_list(self, filename_list):
        return {x: self._query(x) for x in filename_list}

    def _retry_cleanup(self):
        pass

    def _close(self):
        pass

    def _error_code(self, operation, e):  # pylint: disable=unused-argument
        return log.ErrorCode.backend_error

    def __hash_file(self, filename):
        self.__hash_fileobj(open(filename, "rb"))

    def __hash_fileobj(self, fileobj):
        h = hashlib.sha1(usedforsecurity=False)
        # loop till the end of the file
        chunk = 0
        while chunk != b"":
            # read only 1024 bytes at a time
            chunk = fileobj.read(1024)
            h.update(chunk)
        fileobj.close()
        # return the hex representation of digest
        return h.hexdigest()


duplicity.backend.register_backend("fortestsonly", _TestBackend)



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4; coding:utf-8 -*-
#
# Copyright 2002 Ben Escoto
# Copyright 2007 Kenneth Loafman
# Copyright 2019 Carl A. Adams
#
# This file is part of duplicity.
#
# Duplicity is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# Duplicity is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with duplicity; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

"""
S3 Backend using boto3.
"""

import os

import duplicity.backend
from duplicity import config
from duplicity import file_naming
from duplicity import errors
from duplicity import log
from duplicity import progress
from duplicity.errors import BackendException
from duplicity.errors import FatalBackendException

global boto3, botocore, ClientError, S3UploadFailedError, TransferConfig


# Note: current gaps with the old boto backend include:
#       - Glacier restore to S3 not implemented. Should this
#         be done here? Or is that out of scope. My current opinion
#         is that it is out of scope, and the manpage reflects this.
#         It can take days, so waiting seems like it's not ideal.
#         "Thaw" isn't currently a generic concept that the core asks
#         of back-ends. Perhaps that is worth exploring.  The older
#         boto backend appeared  to attempt this restore in the code,
#         but the man page indicated that restores should be done out
#         of band. If implemented,  We should add the the following
#         new features:
#              - when restoring from glacier or deep archive, specify TTL.
#              - allow user to specify how fast to restore (impacts cost).


class S3Boto3Backend(duplicity.backend.Backend):
    """
    Backend for Amazon's Simple Storage System, (aka Amazon S3), though
    the use of the boto3 module. (See
    https://boto3.amazonaws.com/v1/documentation/api/latest/index.html
    for information on boto3.)

    Pursuant to Amazon's announced deprecation of path style S3 access,
    this backend only supports virtual host style bucket URIs.
    See the man page for full details.

    To make use of this backend, you must provide AWS credentials.
    This may be done in several ways: through the environment variables
    AWS_ACCESS_KEY_ID and AWS_SECRET_ACCESS_KEY, by the
    ~/.aws/credentials file, by the ~/.aws/config file,
    or by using the boto2 style ~/.boto or /etc/boto.cfg files.
    """

    def __init__(self, parsed_url):
        """
        Initialize the S3 boto3 backend.

        :param ParsedUrl parsed_url: The parsed URL for the backend.
        :raises BackendException: If the bucket name is missing.
        """
        global boto3, botocore, ClientError, S3UploadFailedError, TransferConfig
        import boto3
        import botocore
        from botocore.exceptions import ClientError
        from boto3.exceptions import S3UploadFailedError
        from boto3.s3.transfer import TransferConfig

        duplicity.backend.Backend.__init__(self, parsed_url)

        # This folds the null prefix and all null parts, which means that:
        #  //MyBucket/ and //MyBucket are equivalent.
        #  //MyBucket//My///My/Prefix/ and //MyBucket/My/Prefix are equivalent.
        url_path_parts = [x for x in parsed_url.path.split("/") if x != ""]
        if url_path_parts:
            self.bucket_name = url_path_parts.pop(0)
        else:
            raise BackendException("S3 requires a bucket name.")

        if url_path_parts:
            self.key_prefix = f"{'/'.join(url_path_parts)}/"
        else:
            self.key_prefix = ""

        self.parsed_url = parsed_url
        self.straight_url = duplicity.backend.strip_auth_from_url(parsed_url)
        self.s3 = None
        self.bucket = None
        self.tracker = UploadProgressTracker()

        if not (boto3.__version__ < "1.36.0" and botocore.__version__ < "1.36.0"):
            # this is an issue with 3rd party s3 implementations only
            # likely when an endpoint is given that resides not under amazonaws.com
            # in time that workaround will probably not be needed anymore
            # https://github.com/boto/boto3/issues/2913
            import re

            if config.s3_endpoint_url and not re.match(
                pattern="(?i).*\\.amazonaws\\.com(/+)?$", string=config.s3_endpoint_url
            ):
                log.Notice(
                    "WARNING: If boto3 >= 1,36.0 with non-amazon s3 services"
                    " results in checksum errors preventing uploads you may try"
                    " the following workaround. set the following env vars\n\n"
                    "    export AWS_REQUEST_CHECKSUM_CALCULATION=when_required\n"
                    "    export AWS_RESPONSE_CHECKSUM_VALIDATION=when_required\n\n"
                    "see https://gitlab.com/duplicity/duplicity/-/issues/870 for details."
                )

    def reset_connection(self):
        """
        Reset the S3 connection.
        """
        self.bucket = None
        self.s3 = boto3.resource(
            "s3",
            region_name=config.s3_region_name,
            use_ssl=(not config.s3_unencrypted_connection),
            endpoint_url=config.s3_endpoint_url,
        )

        try:
            self.s3.meta.client.head_bucket(Bucket=self.bucket_name)
        except botocore.exceptions.ClientError as bce:
            error_code = bce.response["Error"]["Code"]
            if error_code == "404":
                raise FatalBackendException(
                    f'S3 bucket "{self.bucket_name}" does not exist',
                    code=log.ErrorCode.backend_not_found,
                )
            else:
                raise

        self.bucket = self.s3.Bucket(self.bucket_name)  # only set if bucket is thought to exist.

    def _put(self, local_source_path, remote_filename):
        """
        Copy a local file to the remote S3 bucket.
        """
        if not self.s3:
            self.reset_connection()

        # files that should not in glacier and deep_archive, to allow smooth operation
        glacier_exceptions = [
            file_naming.full_manifest_re,
            file_naming.inc_manifest_re,
            file_naming.full_sig_re,
            file_naming.new_sig_re,
        ]

        def is_glacier_exception(filename):
            return any([x.match(filename) for x in glacier_exceptions])

        if config.s3_use_rrs:
            storage_class = "REDUCED_REDUNDANCY"
        elif config.s3_use_ia:
            storage_class = "STANDARD_IA"
        elif config.s3_use_onezone_ia:
            storage_class = "ONEZONE_IA"
        elif config.s3_use_glacier and not is_glacier_exception(remote_filename):
            storage_class = "GLACIER"
        elif config.s3_use_glacier_ir and not is_glacier_exception(remote_filename):
            storage_class = "GLACIER_IR"
        elif config.s3_use_deep_archive and not is_glacier_exception(remote_filename):
            storage_class = "DEEP_ARCHIVE"
        else:
            storage_class = "STANDARD"
        extra_args = {"StorageClass": storage_class}

        if config.s3_use_sse:
            extra_args["ServerSideEncryption"] = "AES256"
        elif config.s3_use_sse_kms:
            if config.s3_kms_key_id is None:
                raise FatalBackendException(
                    "S3 USE SSE KMS was requested, but key id not provided " "require (--s3-kms-key-id)",
                    code=log.ErrorCode.s3_kms_no_id,
                )
            extra_args["ServerSideEncryption"] = "aws:kms"
            extra_args["SSEKMSKeyId"] = config.s3_kms_key_id
            if config.s3_kms_grant:
                extra_args["GrantFullControl"] = config.s3_kms_grant

        transfer_config = TransferConfig(
            multipart_chunksize=config.s3_multipart_chunk_size,
            multipart_threshold=config.s3_multipart_chunk_size,
            max_concurrency=config.s3_multipart_max_procs,
        )

        # Should the tracker be scoped to the put or the backend?
        # The put seems right to me, but the results look a little more correct
        # scoped to the backend.  This brings up questions about knowing when
        # it's proper for it to be reset.
        # tracker = UploadProgressTracker() # Scope the tracker to the put()
        tracker = self.tracker

        remote_filename = os.fsdecode(remote_filename)
        key = self.key_prefix + remote_filename

        log.Info(f"Uploading {self.straight_url}/{remote_filename} to {storage_class} Storage")
        try:
            self.s3.Object(self.bucket.name, key).upload_file(
                local_source_path.uc_name,
                Callback=tracker.progress_cb,
                Config=transfer_config,
                ExtraArgs=extra_args,
            )
        except S3UploadFailedError as e:
            if boto3.__version__ > "1.36.0" or botocore.__version__ > "1.36.0":
                raise FatalBackendException(
                    f"Failed to upload file {remote_filename}, got S3UploadFailedError\n"
                    f"See https://gitlab.com/duplicity/duplicity/-/issues/870 for details.\n"
                    f"Quick fix is: [sudo] pip3 install boto3<1.36.0 botocore<1.36.0\n"
                    f"API-Error: {e}"
                )
            else:
                raise e

    def _get(self, remote_filename, local_path):
        """
        Copy a remote S3 object to a local path.

        :param str remote_filename: The name of the file on the backend.
        :param Path local_path: The local path to copy the file to.
        :raises FatalError: If the file is in long term storage (Glacier).
        """
        if not self.s3:
            self.reset_connection()

        remote_filename = os.fsdecode(remote_filename)
        key = self.key_prefix + remote_filename
        try:
            self.s3.Object(self.bucket.name, key).download_file(local_path.uc_name)
        except ClientError as ios:
            if ios.response["Error"]["Code"] == "InvalidObjectState":
                raise errors.FatalError(
                    f"File {remote_filename} seems to be in a long term storage,"
                    f"Please use AWS Console/API to initiate restore.\n"
                    f"API-Error: {ios}"
                )
            else:
                raise ios

    def _list(self):
        """
        List objects in the S3 bucket under the key prefix.

        :returns: A list of filenames as byte strings.
        :rtype: list
        """
        if not self.s3:
            self.reset_connection()

        filename_list = []
        for obj in self.bucket.objects.filter(Prefix=self.key_prefix):
            try:
                filename = obj.key.replace(self.key_prefix, "", 1)
                filename_list.append(os.fsencode(filename))
                log.Debug(f"Listed {self.straight_url}/{filename}")
            except AttributeError:
                pass
        return filename_list

    def _delete(self, remote_filename):
        """
        Delete an object from the S3 bucket.

        :param str remote_filename: The name of the file to delete.
        """
        if not self.s3:
            self.reset_connection()

        remote_filename = os.fsdecode(remote_filename)
        key = self.key_prefix + remote_filename
        self.s3.Object(self.bucket.name, key).delete()

    def _query(self, remote_filename):
        """
        Query metadata for an object in the S3 bucket.

        :param str remote_filename: The name of the object to query.
        :returns: A dictionary containing metadata (e.g., 'size').
        :rtype: dict
        """
        if not self.s3:
            self.reset_connection()

        remote_filename = os.fsdecode(remote_filename)
        key = self.key_prefix + remote_filename
        content_length = -1
        try:
            s3_obj = self.s3.Object(self.bucket.name, key)
            s3_obj.load()
            content_length = s3_obj.content_length
        except botocore.exceptions.ClientError as bce:
            if bce.response["Error"]["Code"] == "404":
                pass
            else:
                raise
        return {"size": content_length}


class UploadProgressTracker(object):
    """
    Track progress of an upload.
    """

    def __init__(self):
        """
        Initialize the upload progress tracker.
        """
        self.total_bytes = 0

    def progress_cb(self, fresh_byte_count):
        """
        Callback for boto3 to report progress.
        """
        self.total_bytes += fresh_byte_count
        progress.report_transfer(self.total_bytes, 0)  # second arg appears to be unused
        # It would seem to me that summing progress should be the callers job,
        # and backends should just toss bytes written numbers over the fence.
        # But, the progress bar doesn't work in a reasonable way when we do
        # that. (This would also eliminate the need for this class to hold
        # the scoped rolling total.)
        # progress.report_transfer(fresh_byte_count, 0)


duplicity.backend.register_backend("boto3+s3", S3Boto3Backend)
# make boto3 the default s3 backend
duplicity.backend.register_backend("s3", S3Boto3Backend)



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4; coding:utf-8 -*-
#
# Copyright 2009 Eric EJ Johnson
#
# This file is part of duplicity.
#
# Duplicity is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# Duplicity is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with duplicity; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

"""
Rackspace CloudFiles backend (legacy implementation).
"""

import os

import duplicity.backend
from duplicity import errors
from duplicity import log
from duplicity import util
from duplicity.errors import BackendException


class CloudFilesBackend(duplicity.backend.Backend):
    """
    Backend for Rackspace's CloudFiles
    """

    def __init__(self, parsed_url):
        """
        Initialize the CloudFiles backend.

        :param ParsedUrl parsed_url: The parsed URL.
        """
        super().__init__(parsed_url)
        try:
            from cloudfiles import Connection
            from cloudfiles.errors import ResponseError
            from cloudfiles import consts
            from cloudfiles.errors import NoSuchObject
        except ImportError as e:
            raise BackendException(f"""Cloudfiles backend requires the cloudfiles library available from Rackspace.
Exception: {str(e)}""")

        self.resp_exc = ResponseError
        conn_kwargs = {}

        if "CLOUDFILES_USERNAME" not in os.environ:
            raise BackendException("CLOUDFILES_USERNAME environment variable" "not set.")

        if "CLOUDFILES_APIKEY" not in os.environ:
            raise BackendException("CLOUDFILES_APIKEY environment variable not set.")

        conn_kwargs["username"] = os.environ["CLOUDFILES_USERNAME"]
        conn_kwargs["api_key"] = os.environ["CLOUDFILES_APIKEY"]

        if "CLOUDFILES_AUTHURL" in os.environ:
            conn_kwargs["authurl"] = os.environ["CLOUDFILES_AUTHURL"]
        else:
            conn_kwargs["authurl"] = consts.default_authurl

        container = parsed_url.path.lstrip("/")

        try:
            conn = Connection(**conn_kwargs)
        except Exception as e:
            raise errors.FatalError(
                f"Connection failed, please check your credentials: {e.__class__.__name__} {util.uexc(e)}",
                log.ErrorCode.connection_failed,
            )
        self.container = conn.create_container(container)

    def _error_code(self, operation, e):  # pylint: disable=unused-argument
        if isinstance(e, NoSuchObject):
            return log.ErrorCode.backend_not_found
        elif isinstance(e, self.resp_exc):
            if e.status == 404:
                return log.ErrorCode.backend_not_found

    def _put(self, source_path, remote_filename):
        sobject = self.container.create_object(remote_filename)
        sobject.load_from_filename(source_path.name)

    def _get(self, remote_filename, local_path):
        sobject = self.container.create_object(remote_filename)
        with open(local_path.name, "wb") as f:
            for chunk in sobject.stream():
                f.write(chunk)

    def _list(self):
        """
        List files in the CloudFiles container.

        Handles the 10,000 object limit by making multiple requests.

        :returns: A list of filenames.
        :rtype: list
        """
        # Cloud Files will return a max of 10,000 objects.  We have
        # to make multiple requests to get them all.
        objs = self.container.list_objects()
        keys = objs
        while len(objs) == 10000:
            objs = self.container.list_objects(marker=keys[-1])
            keys += objs
        return keys

    def _delete(self, filename):
        self.container.delete_object(filename)

    def _query(self, filename):
        sobject = self.container.get_object(filename)
        return {"size": sobject.size}



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4; coding:utf-8 -*-
#
# Copyright 2002 Ben Escoto
# Copyright 2007 Kenneth Loafman
#
# This file is part of duplicity.
#
# Duplicity is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# Duplicity is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with duplicity; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

"""
Miscellaneous utilities.
"""

import atexit
import csv
import errno
import json
import multiprocessing
import os
import socket
import sys
import traceback
from io import StringIO

import fasteners

from duplicity import config
from duplicity import dup_tarfile
from duplicity import errors
from duplicity import log


def exception_traceback(limit=50):
    """
    Return a string representation of the currently active/raised exception.

    The output format matches the typical Python traceback format.

    :param int limit: The maximum number of traceback levels to include.
    :returns: A string containing the formatted traceback.
    :rtype: str
    """
    type, value, tb = sys.exc_info()  # pylint: disable=redefined-builtin

    lines = traceback.format_tb(tb, limit)
    lines.extend(traceback.format_exception_only(type, value))

    msg = "Traceback (innermost last):\n"
    msg = f"{msg}{str.join('', lines[:-1]):20} {lines[-1]}"

    return msg


def escape(string):
    """
    Convert a filename to a format suitable for logging.

    The filename is quoted and escaped using UTF-8.

    :param bytes or str string: The filename to escape.
    :returns: The escaped filename string.
    :rtype: str
    """
    string = os.fsdecode(string).encode("unicode-escape", "replace")
    return "'%s'" % string.decode("utf8", "replace").replace("'", "\\x27")


def uindex(index):
    """
    Convert a path index to a string for printing.

    :param tuple index: The path index (a tuple of path parts).
    :returns: The path string.
    :rtype: str
    """
    if index:
        return os.path.join(*list(map(os.fsdecode, index)))
    else:
        return "."


def uexc(e):
    """
    Return the exception message as a Unicode string.

    Handles decoding of byte-string arguments using the filesystem encoding.

    :param Exception e: The exception object.
    :returns: The exception message.
    :rtype: str
    """
    # Exceptions in duplicity often have path names in them, which if they are
    # non-ascii will cause a UnicodeDecodeError when implicitly decoding to
    # unicode.  So we decode manually, using the filesystem encoding.
    # 99.99% of the time, this will be a fine encoding to use.
    if e and e.args:
        # Find arg that is a string
        for m in e.args:
            if isinstance(m, str):
                # Already unicode
                return m
            elif isinstance(m, bytes):
                # Encoded, likely in filesystem encoding
                return os.fsdecode(m)
        # If the function did not return yet, we did not
        # succeed in finding a string; return the whole message.
        return str(e)
    else:
        return ""


def maybe_ignore_errors(fn):
    """
    Execute ``fn``, optionally ignoring errors.

    If the global configuration setting ``ignore_errors`` is set to True,
    catch errors and log them but continue execution.

    :param callable fn: The callable to execute.
    :returns: The result of ``fn()``, or None if it failed and ``ignore_errors`` is true.
    """
    try:
        return fn()
    except Exception as e:
        if config.ignore_errors:
            log.Warn(_("IGNORED_ERROR: WARNING: ignoring error as requested: %s: %s") % (e.__class__.__name__, uexc(e)))
            return None
        else:
            raise


class BlackHoleList(list):
    """
    A list that discards everything appended to it.
    """

    def append(self, x):
        """
        Discard the given element.
        """
        pass


class FakeTarFile(object):
    """
    A fake TarFile object that does nothing.
    """

    debug = 0

    def __iter__(self):
        """
        Return an empty iterator.
        """
        return iter([])

    def close(self):
        """
        Do nothing on close.
        """
        pass


def make_tarfile(mode, fp):
    """
    Create a TarFile object with the given mode and file pointer.

    :param str mode: The mode to open the tarfile in.
    :param fp: The file pointer.
    :returns: A TarFile object or a FakeTarFile if it is empty.
    """
    # We often use 'empty' tarfiles for signatures that haven't been filled out
    # yet.  So we want to ignore ReadError exceptions, which are used to signal
    # this.
    try:
        tf = dup_tarfile.TarFile("arbitrary", mode, fp)
        # Now we cause TarFile to not cache TarInfo objects.  It would end up
        # consuming a lot of memory over the lifetime of our long-lasting
        # signature files otherwise.
        tf.members = BlackHoleList()
        return tf
    except dup_tarfile.ReadError:
        return FakeTarFile()


def get_tarinfo_name(ti):
    """
    Return the name of a TarInfo object, ensuring it ends with / if it is a directory.
    """
    # Python versions before 2.6 ensure that directories end with /, but 2.6
    # and later ensure they they *don't* have /.  ::shrug::  Internally, we
    # continue to use pre-2.6 method.
    if ti.isdir() and not ti.name.endswith(r"/"):
        return f"{ti.name}/"
    else:
        return ti.name


def ignore_missing(fn, filename):
    """
    Execute ``fn`` on ``filename``, ignoring ``ENOENT`` errors.

    :param callable fn: The function to execute.
    :param str filename: The filename to pass to the function.
    :raises OSError: If an error other than ``ENOENT`` occurs.
    """
    try:
        fn(filename)
    except OSError as ex:
        if ex.errno == errno.ENOENT:
            pass
        else:
            raise


def acquire_lockfile():
    """
    Acquire the lockfile if it doesn't exist.
    """
    config.lockpath = os.path.join(config.archive_dir_path.name, b"lockfile")
    config.lockfile = fasteners.process_lock.InterProcessLock(config.lockpath)
    log.Log(
        _("Acquiring lockfile %s") % os.fsdecode(config.lockpath),
        log.DEBUG,
    )
    if not config.lockfile.acquire(blocking=False):
        raise errors.FatalError(
            f"Another duplicity instance is already running with this archive directory\n"
            f"If this is not the case, remove {config.lockpath}'.",
            log.ErrorCode.user_error,
        )


@atexit.register
def release_lockfile():
    """
    Release the lockfile.  Don't do anything if it doesn't exist.
    """
    if config.lockfile is not None:
        try:
            config.lockfile.release()
        except Exception as e:
            log.Error(f"Could not release lockfile: {str(e)}")
            pass
        try:
            os.unlink(config.lockpath)
        except Exception as e:
            log.Error(f"Could not delete lockfile: {str(e)}")
            pass
        config.lockfile = None
        config.lockpath = ""


def copyfileobj(infp, outfp, byte_count=-1):
    """
    Copy byte_count bytes from infp to outfp, or all if byte_count < 0

    Returns the number of bytes actually written (may be less than
    byte_count if find eof.  Does not close either fileobj.

    """
    blocksize = 64 * 1024
    bytes_written = 0
    if byte_count < 0:
        while True:
            buf = infp.read(blocksize)
            if not buf:
                break
            bytes_written += len(buf)
            outfp.write(buf)
    else:
        while bytes_written + blocksize <= byte_count:
            buf = infp.read(blocksize)
            if not buf:
                break
            bytes_written += len(buf)
            outfp.write(buf)
        buf = infp.read(byte_count - bytes_written)
        bytes_written += len(buf)
        outfp.write(buf)
    return bytes_written


def which(program):
    """
    Return the absolute path for the given program name.

    :param str program: The name of the program to find.
    :returns: The absolute path to the program, or None if not found.
    :rtype: str or None
    """

    def is_exe(fpath):
        return os.path.isfile(fpath) and os.path.isabs(fpath) and os.access(fpath, os.X_OK)

    fpath, fname = os.path.split(program)
    if fpath:
        if is_exe(program):
            return program
    else:
        for path in os.getenv("PATH").split(os.pathsep):
            path = path.strip('"')
            exe_file = os.path.abspath(os.path.join(path, program))
            if is_exe(exe_file):
                return exe_file

    return None


def start_debugger():
    """
    Start the debugger if requested by the environment.
    """
    import socket
    import errno

    def is_port_in_use(host: str, port: int) -> bool:
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
            try:
                # Bind to host on the specified port
                s.bind((host, port))
                return False  # Port is free
            except socket.error as e:
                if e.errno == errno.EADDRINUSE:
                    return True  # Port is already in use
                else:
                    # Something else went wrong
                    raise e

    def fix_path():
        """
        in a dev environment the path is screwed so fix it.
        """
        base = sys.path.pop(0)
        base = base.split(os.path.sep)[:-1]
        base = os.path.sep.join(base)
        sys.path.insert(0, base)

    if os.environ.get("PYDEVD", None) == "vscode":
        try:
            import debugpy  # pylint: disable=import-error
        except ImportError:
            raise errors.FatalError(
                "Module debugpy must be available for debugging.\n"
                "Don't set 'PYDEVD=vscode'\n"
                "to avoid starting debugpy as debugger."
            )
        host = "127.0.0.1"
        pool_nr = 0
        if "multiprocessing" in sys.modules:
            try:
                pool_nr = multiprocessing.current_process()._identity[0]
                print(f"Debugger for Pool Member: {pool_nr}")
            except IndexError:
                print("Debugger not in pool process.")
        port = 5678 + pool_nr

        while is_port_in_use(host, port):
            port += 1
        print(f"Debugger start on port {port}")
        debugpy.listen(port)
        print(f"Waiting for debugger attach on port: {port}")
        debugpy.wait_for_client()
        fix_path()

    elif config.pydevd or os.environ.get("PYDEVD", None):
        try:
            import debugpy  # pylint: disable=import-error
        except ImportError:
            raise errors.FatalError(
                "Module debugpy must be available for debugging.\n"
                "Remove '--pydevd' from command line and unset 'PYDEVD'\n"
                "from the environment to avoid starting the debugger."
            )

        debug_host = "127.0.0.1"
        debug_port = 5678

        # ignition
        print(f"Debugger start on {debug_host}:{debug_port}", file=sys.stderr)
        debugpy.listen([debug_host, debug_port])
        print(f"Waiting for debugger attach on port: {debug_port}", file=sys.stderr)
        debugpy.wait_for_client()
        fix_path()


def merge_dicts(*dict_args):
    """
    Merge multiple dictionaries into a new one.

    Shallow copies each dictionary. Precedence is given to key-value pairs
    in later dictionaries.

    :param dict dict_args: The dictionaries to merge.
    :returns: A new merged dictionary.
    :rtype: dict
    """
    result = {}
    for dictionary in dict_args:
        result.update(dictionary)
    return result


def csv_args_to_dict(arg):
    """
    Given the string arg in single line csv format, split into pairs (key, val)
    and produce a dictionary from those key:val pairs.
    """
    mydict = {}
    with StringIO(arg) as infile:
        rows = csv.reader(infile)
        for row in rows:
            for i in range(0, len(row), 2):
                mydict[row[i]] = row[i + 1]
    return mydict


class BytesEncoder(json.JSONEncoder):
    """
    JSON doesn't allow byte type values. Converting them to Unicode strings
    """

    def default(self, obj):
        """
        Handle byte types by converting them to Unicode strings.
        """
        if isinstance(obj, bytes):
            return obj.decode()
        return json.JSONEncoder.default(self, obj)



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4; coding:utf-8 -*-
#
# duplicity -- Encrypted bandwidth efficient backup
#
# Copyright 2002 Ben Escoto
# Copyright 2007 Kenneth Loafman
#
# This file is part of duplicity.
#
# Duplicity is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# Duplicity is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with duplicity; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
#
# For more information see
#  http://duplicity.us
#  or
#  http://duplicity.gitlab.io
# .
# Please send mail to me or the mailing list if you find bugs or have
# any suggestions.

"""
Main orchestration logic for Duplicity.
"""

import os
import platform
import resource
import sys
import time

from dataclasses import dataclass
from textwrap import dedent
from typing import Dict

from duplicity import __version__
from duplicity import backend
from duplicity import backend_pool
from duplicity import cli_main
from duplicity import config
from duplicity import diffdir
from duplicity import dup_collections
from duplicity import dup_temp
from duplicity import dup_time
from duplicity import file_naming
from duplicity import gpg
from duplicity import errors
from duplicity import log
from duplicity import manifest
from duplicity import patchdir
from duplicity import path
from duplicity import progress
from duplicity import selection
from duplicity import tempdir
from duplicity import util

# If exit_val is not None, exit with given value at end.
exit_val = None

# actions that skip archive_sync
skips_sync_archive = [
    "collection-status",
    "full",
    "remove-all-but-n-full",
    "remove-all-inc-of-but-n-full",
    "remove-older-than",
]


def getpass_safe(message):
    """
    Prompt for a passphrase without echoing input.
    """
    import getpass

    return getpass.getpass(message)


def get_passphrase(n, action, for_signing=False):
    """
    Check to make sure passphrase is indeed needed, then get
    the passphrase from environment, from gpg-agent, or user

    If n=3, a password is requested and verified. If n=2, the current
    password is verified. If n=1, a password is requested without
    verification for the time being.

    :param int n: verification level for a passphrase being requested
    :param str action: action to perform
    :param bool for_signing: true if the passphrase is for a signing key, false if not
    :returns: passphrase
    :rtype: str
    """

    # First try the environment
    try:
        if for_signing:
            return os.environ["SIGN_PASSPHRASE"]
        else:
            return os.environ["PASSPHRASE"]
    except KeyError:
        pass

    # check if we can reuse an already set (signing_)passphrase
    # if signing key is also an encryption key assume that the passphrase is identical
    if (
        for_signing
        and (
            config.gpg_profile.sign_key in config.gpg_profile.recipients
            or config.gpg_profile.sign_key in config.gpg_profile.hidden_recipients
        )
        and "PASSPHRASE" in os.environ
    ):  # noqa
        log.Notice(_("Reuse configured PASSPHRASE as SIGN_PASSPHRASE"))
        return os.environ["PASSPHRASE"]
    # if one encryption key is also the signing key assume that the passphrase is identical
    if (
        not for_signing
        and (
            config.gpg_profile.sign_key in config.gpg_profile.recipients
            or config.gpg_profile.sign_key in config.gpg_profile.hidden_recipients
        )
        and "SIGN_PASSPHRASE" in os.environ
    ):  # noqa
        log.Notice(_("Reuse configured SIGN_PASSPHRASE as PASSPHRASE"))
        return os.environ["SIGN_PASSPHRASE"]

    # Next, verify we need to ask the user

    # Assumptions:
    #   - encrypt-key has no passphrase
    #   - sign-key requires passphrase
    #   - gpg-agent supplies all, no user interaction

    # no passphrase if --no-encryption or --use-agent
    if not config.encryption or config.use_agent:
        return ""

    # these commands don't need a password
    elif action in [
        "collection-status",
        "remove-all-but-n-full",
        "remove-all-inc-of-but-n-full",
        "remove-older-than",
        "cleanup",
    ]:
        return ""

    # for a full, inc, sync, verify or list-current-files, we don't need a password if
    # there is no sign_key and there are recipients
    elif (
        action in ("full", "inc", "sync", "verify", "list-current-files")
        and (config.gpg_profile.recipients or config.gpg_profile.hidden_recipients)
        and (not config.gpg_profile.sign_key or not for_signing)
    ):
        return ""

    elif (
        (config.gpg_profile.recipients or config.gpg_profile.hidden_recipients)
        and config.metadata_sync_mode == "partial"
        and action in ["full"]
    ):
        log.Info(_("Skipping passphrase input for full backup with encryption keys."))
        return ""

    # Finally, ask the user for the passphrase
    else:
        log.Info(_("PASSPHRASE variable not set, asking user."))
        use_cache = True
        while True:
            # ask the user to enter a new passphrase to avoid an infinite loop
            # if the user made a typo in the first passphrase
            if use_cache and n == 2:
                if for_signing:
                    pass1 = config.gpg_profile.signing_passphrase
                else:
                    pass1 = config.gpg_profile.passphrase
            else:
                if for_signing:
                    if use_cache and config.gpg_profile.signing_passphrase is not None:
                        pass1 = config.gpg_profile.signing_passphrase
                    else:
                        pass1 = getpass_safe(f"{_('GnuPG passphrase for signing key:')} ")
                else:
                    if use_cache and config.gpg_profile.passphrase is not None:
                        pass1 = config.gpg_profile.passphrase
                    else:
                        pass1 = getpass_safe(f"{_('GnuPG passphrase for decryption:')} ")

            if n == 1:
                pass2 = pass1
            elif for_signing:
                pass2 = getpass_safe(_("Retype passphrase for signing key to confirm: "))
            else:
                pass2 = getpass_safe(_("Retype passphrase for decryption to confirm: "))

            if not pass1 == pass2:
                log.Log(
                    _("First and second passphrases do not match!  Please try again."),
                    log.WARNING,
                    force_print=True,
                )
                use_cache = False
                continue

            if (
                not pass1
                and not (config.gpg_profile.recipients or config.gpg_profile.hidden_recipients)
                and not for_signing
            ):
                log.Log(
                    _("Cannot use empty passphrase with symmetric encryption!  Please try again."),
                    log.WARNING,
                    force_print=True,
                )
                use_cache = False
                continue

            return pass1


def dummy_backup(tarblock_iter):
    """
    Fake writing to backend, but do go through all the source paths.

    :param tarblock_iter: iterator for current tar block
    :returns: constant 0 (zero)
    :rtype: int
    """
    try:
        # Just spin our wheels
        while next(tarblock_iter):
            pass
    except StopIteration:
        pass
    log.Progress(None, diffdir.stats.SourceFileSize)
    return 0


def restart_position_iterator(tarblock_iter):
    """
    Fake writing to backend, but do go through all the source paths.

    Stop when we have processed the last file and block from the
    last backup.  Normal backup will proceed at the start of the
    next volume in the set.

    :param tarblock_iter: iterator for current tar block
    :returns: constant 0 (zero)
    :rtype: int
    """
    last_index = config.restart.last_index
    last_block = config.restart.last_block
    try:
        # Just spin our wheels
        iter_result = next(tarblock_iter)
        while iter_result:
            if tarblock_iter.previous_index == last_index:
                # If both the previous index and this index are done, exit now
                # before we hit the next index, to prevent skipping its first
                # block.
                if not last_block and not tarblock_iter.previous_block:
                    break
                # Only check block number if last_block is also a number
                if last_block and tarblock_iter.previous_block and tarblock_iter.previous_block > last_block:
                    break
            if tarblock_iter.previous_index > last_index:
                log.Warn(
                    _("File %s complete in backup set.\n" "Continuing restart on file %s.")
                    % (util.uindex(last_index), util.uindex(tarblock_iter.previous_index)),
                    log.ErrorCode.restart_file_not_found,
                )
                # We went too far! Stuff the data back into place before restarting
                tarblock_iter.queue_index_data(iter_result)
                break
            iter_result = next(tarblock_iter)
    except StopIteration:
        log.Warn(
            _("File %s missing in backup set.\n" "Continuing restart on file %s.")
            % (util.uindex(last_index), util.uindex(tarblock_iter.previous_index)),
            log.ErrorCode.restart_file_not_found,
        )


def write_multivol(backup_type, tarblock_iter, man_outfp, sig_outfp, backend):
    """
    Encrypt volumes of tarblock_iter and write to backend.

    backup_type should be "inc" or "full" and only matters here when
    picking the filenames.  The path_prefix will determine the names
    of the files written to backend.  Also writes manifest file.

    :param str backup_type: type of backup to perform, either 'inc' or 'full'
    :param tarblock_iter: iterator for current tar block
    :param backend: I/O backend for selected protocol
    :returns: bytes written
    :rtype: int
    """

    def get_indicies(tarblock_iter):
        """
        Return start_index and end_index of previous volume
        """
        start_index, start_block = tarblock_iter.recall_index()
        if start_index is None:
            start_index = ()
            start_block = None
        if start_block:
            start_block -= 1
        end_index, end_block = tarblock_iter.get_previous_index()
        if end_index is None:
            end_index = start_index
            end_block = start_block
        if end_block:
            end_block -= 1
        return start_index, start_block, end_index, end_block

    def put(tdp, dest_filename, vol_num):
        """
        Retrieve file size *before* calling backend.put(), which may (at least
        in case of the localbackend) rename the temporary file to the target
        instead of copying.
        """
        putsize = tdp.getsize()
        if config.skip_volume != vol_num:  # for testing purposes only
            backend.put(tdp, dest_filename)
        res, msg = backend.validate(dest_filename, putsize, source_path=tdp)
        if not res:
            code_extra = f"{util.escape(dest_filename)}: {msg}"
            raise errors.FatalError(
                _("File %s was corrupted during upload.") % os.fsdecode(dest_filename),
                log.ErrorCode.backend_validation_failed,
            )

        if tdp.stat:
            tdp.delete()
        return putsize

    def validate_encryption_settings(backup_set, manifest):
        """
        When restarting a backup, we have no way to verify that the current
        passphrase is the same as the one used for the beginning of the backup.
        This is because the local copy of the manifest is unencrypted and we
        don't need to decrypt the existing volumes on the backend.  To ensure
        that we are using the same passphrase, we manually download volume 1
        and decrypt it with the current passphrase.  We also want to confirm
        that we're using the same encryption settings (i.e. we don't switch
        from encrypted to non in the middle of a backup chain), so we check
        that the vol1 filename on the server matches the settings of this run.
        """
        if (config.gpg_profile.recipients or config.gpg_profile.hidden_recipients) and not config.gpg_profile.sign_key:
            # When using gpg encryption without a signing key, we skip this validation
            # step to ensure that we can still backup without needing the secret key
            # on the machine.
            return

        vol1_filename = file_naming.get(backup_type, 1, encrypted=config.encryption, gzipped=config.compression)
        if vol1_filename != backup_set.volume_name_dict[1]:
            raise errors.FatalError(
                _("Restarting backup, but current encryption " "settings do not match original settings"),
                log.ErrorCode.enryption_mismatch,
            )

        # Settings are same, let's check passphrase itself if we are encrypted
        if config.encryption:
            fileobj = restore_get_enc_fileobj(config.backend, vol1_filename, manifest.volume_info_dict[1])
            fileobj.close()

    @dataclass
    class CommandMetaData:
        """
        Track per-volume state for a queued concurrent backend upload.

        Records the volume number, temp path, and volume info, plus flags
        indicating whether the transfer succeeded and whether its info has
        been written to the manifest.
        """

        vol_num: int
        path_obj: path.Path
        vol_info: manifest.VolumeInfo
        transfer_success = False
        manifest_written = False

    def collect_put_results(backend_pooler, command2vol_map: Dict[int, CommandMetaData]):
        """
        Collect completed upload results and update progress state.
        """
        bytes_written = 0
        for result in backend_pooler.results_since_last_call():
            track_id = result.track_id
            size = result.result
            bytes_written += size
            progress.report_transfer(size, size)
            log.Progress(_("Processed volume %d") % command2vol_map[track_id].vol_num, bytes_written)
            command2vol_map[track_id].transfer_success = True
            if command2vol_map[track_id].path_obj.stat:
                command2vol_map[track_id].path_obj.delete()
            if config.progress:
                progress.tracker.snapshot_progress(command2vol_map[track_id].vol_num)
            log.Debug(
                f"Transfer of {command2vol_map[track_id].path_obj.get_filename()} with id {track_id} and size "
                f"{size} took {result.get_runtime()}"
            )
        return bytes_written

    def write_manifest_in_sequence(mf, mf_file, command2vol_map: Dict[int, CommandMetaData]):
        """
        Ensure volume info is written only if the volume transfer was successful and in sequence
        without gap otherwise missing volumes won't be detected and data is missing.
        """
        try:
            i = iter(command2vol_map.keys())
            info = command2vol_map[next(i)]
            while info.manifest_written and info.transfer_success:
                # skip all entries already written
                info = command2vol_map[next(i)]
            while info.transfer_success:
                # if next volumes are transferred
                if not info.manifest_written:
                    mf.add_volume_info(info.vol_info)
                info.manifest_written = True

                if info.vol_num == 1:
                    mf_file.to_partial()
                    sig_outfp.to_partial()
                else:
                    mf_file.flush()
                    sig_outfp.flush()

                info = command2vol_map[next(i)]
        except StopIteration:
            pass  # no manifest to update

    if not config.restart:
        # normal backup start
        vol_num = 0
        mf = manifest.Manifest(fh=man_outfp)
        mf.set_dirinfo()
    else:
        # restart from last known position
        mf = config.restart.last_backup.get_local_manifest()
        config.restart.checkManifest(mf)
        config.restart.setLastSaved(mf)
        if not (config.s3_use_deep_archive or config.s3_use_glacier or config.s3_use_glacier_ir):
            validate_encryption_settings(config.restart.last_backup, mf)
        else:
            log.Warn(_("Skipping encryption validation due to glacier/deep storage"))
        mf.fh = man_outfp
        last_block = config.restart.last_block
        log.Notice(
            _("Restarting after volume %s, file %s, block %s")
            % (config.restart.start_vol, util.uindex(config.restart.last_index), config.restart.last_block)
        )
        vol_num = config.restart.start_vol
        restart_position_iterator(tarblock_iter)

    at_end = 0
    bytes_written = 0

    # If --progress option is given, initiate a background thread that will
    # periodically report progress to the Log.
    if config.progress:
        progress.tracker.set_start_volume(vol_num + 1)
        progress.progress_thread.start()

    backend_pooler = None
    command2vol_map: Dict[int, CommandMetaData] = {}
    if config.concurrency > 0:
        backend_pooler = backend_pool.BackendPool(config.target_url, processes=config.concurrency)

    while not at_end:
        # set up iterator
        tarblock_iter.remember_next_index()  # keep track of start index

        # Create volume
        vol_num += 1
        dest_filename = file_naming.get(backup_type, vol_num, encrypted=config.encryption, gzipped=config.compression)
        tdp = dup_temp.new_tempduppath(file_naming.parse(dest_filename))

        # write volume
        if config.encryption:
            at_end = gpg.GPGWriteFile(tarblock_iter, tdp.name, config.gpg_profile, config.volsize)
        elif config.compression:
            at_end = gpg.GzipWriteFile(tarblock_iter, tdp.name, config.volsize)
        else:
            at_end = gpg.PlainWriteFile(tarblock_iter, tdp.name, config.volsize)
        tdp.setdata()

        # Add volume information to manifest
        vi = manifest.VolumeInfo()
        vi.set_info(vol_num, *get_indicies(tarblock_iter))
        vi.set_hash("SHA1", gpg.get_hash("SHA1", tdp))

        if config.skip_if_no_change and diffdir.stats.DeltaEntries == 0 and at_end and vol_num == 1:
            # if nothing changed, skip upload if configured.
            msg = _("Skipped volume upload, as effectivly nothing has changed")
            log.Progress(msg, diffdir.stats.SourceFileSize)
            log.Notice(msg)
            config.skipped_inc = True
            tdp.delete()
            continue

        if config.concurrency > 0 and backend_pooler:
            try:
                # use process pool the write volumes in the background.
                progress.report_transfer(0, tdp.getsize())
                track_id = backend_pooler.command_throttled(backend.put_validated.__name__, args=(tdp, dest_filename))
                command2vol_map[track_id] = CommandMetaData(vol_num, tdp, vi)
                bytes_written += collect_put_results(backend_pooler, command2vol_map)
                write_manifest_in_sequence(mf, man_outfp, command2vol_map)
            except (Exception, SystemExit) as e:
                # ensure pool processes terminate clean
                backend_pooler.shutdown()
                raise

        else:
            # write Volume to backend - blocking
            bytes_written += put(tdp, dest_filename, vol_num)

            mf.add_volume_info(vi)

            if vol_num == 1:
                man_outfp.to_partial()
                sig_outfp.to_partial()
            else:
                man_outfp.flush()
                sig_outfp.flush()

            # Log human-readable version as well as raw numbers for machine consumers
            log.Progress(_("Processed volume %d") % vol_num, diffdir.stats.SourceFileSize)
            # Snapshot (serialize) progress now as a Volume has been completed.
            # This is always the last restore point when it comes to restart a failed backup
            if config.progress:
                progress.tracker.snapshot_progress(vol_num)

    if backend_pooler:
        try:
            # wait for background commands, collect some stats and shutdown clean.
            log.Debug("Collecting remaining results from backend pool.")
            while True and backend_pooler:
                bytes_written += collect_put_results(backend_pooler, command2vol_map)
                write_manifest_in_sequence(mf, man_outfp, command2vol_map)
                if backend_pooler.get_queue_length() == 0:
                    break
                else:
                    log.Debug(f"Still {backend_pooler.get_queue_length()} commands left, waiting ...")
                    time.sleep(2)
            # print some stats, that can be used to track efficiency of concurrent transfers.
            stats = backend_pooler.get_stats()
            log.Info(
                f'Transferred {stats["count"]} volumes, a volume took avg: {stats["time"]["avg"]}, '
                + f'max: {stats["time"]["max"]}, min: {stats["time"]["min"]}'
            )
            # double check that all volumes are transferred
            if not all([x.transfer_success for x in command2vol_map.values()]) and not config.skipped_inc:
                failed_volume_numbers = [
                    x.vol_info.volume_number for x in command2vol_map.values() if not x.transfer_success
                ]
                raise errors.FatalError(f"Volumes with number {failed_volume_numbers} were not transferred successful.")
            # Add some stats, accessible with `--jsonstats`. Not the most elegant way, but it is working.
            diffdir.stats.stat_attrs += ("ConcurrentTransferStats",)
            diffdir.stats.set_stat("ConcurrentTransferStats", stats)
        finally:
            # make sure pool get terminated gracefully in any case.
            backend_pooler.shutdown()

    # Upload the collection summary.
    # bytes_written += write_manifest(mf, backup_type, backend)
    mf.set_files_changed_info(diffdir.stats.get_delta_entries_file())

    return bytes_written


def get_man_fileobj(backup_type):
    """
    Return a fileobj opened for writing, save results as manifest

    Save manifest in config.archive_dir_path gzipped.
    Save them on the backend encrypted as needed.

    :param str backup_type: type of backup to perform, either 'inc' or 'full'
    :returns: fileobj opened for writing
    :rtype: fileobj
    """
    assert backup_type == "full" or backup_type == "inc", f"backup_type must be 'full' or 'inc', got {backup_type!r}"

    part_man_filename = file_naming.get(backup_type, manifest=True, partial=True)
    perm_man_filename = file_naming.get(backup_type, manifest=True)
    remote_man_filename = file_naming.get(backup_type, manifest=True, encrypted=config.encryption)

    fh = dup_temp.get_fileobj_duppath(
        config.archive_dir_path, part_man_filename, perm_man_filename, remote_man_filename
    )
    return fh


def get_sig_fileobj(sig_type):
    """
    Return a fileobj opened for writing, save results as signature

    Save signatures in config.archive_dir gzipped.
    Save them on the backend encrypted as needed.

    :param str sig_type: either "full-sig" or "new-sig"
    :returns: fileobj opened for writing
    :rtype: fileobj
    """
    assert sig_type in ["full-sig", "new-sig"], f"sig_type must be 'full-sig' or 'new-sig', got {sig_type!r}"

    part_sig_filename = file_naming.get(sig_type, gzipped=False, partial=True)
    perm_sig_filename = file_naming.get(sig_type, gzipped=True)
    remote_sig_filename = file_naming.get(sig_type, encrypted=config.encryption, gzipped=config.compression)

    fh = dup_temp.get_fileobj_duppath(
        config.archive_dir_path, part_sig_filename, perm_sig_filename, remote_sig_filename, overwrite=True
    )
    return fh


def get_stat_fileobj(stat_type):
    """
    Return a fileobj opened for writing, save statistic as json

    Save statistics in config.archive_dir gzipped.
    Save them on the backend encrypted as needed.

    :param str stat_type: either "full-stat" or "new-stat"
    :returns: fileobj opened for writing
    :rtype: fileobj
    """
    assert stat_type in ["full-stat", "inc-stat"], f"stat_type must be 'full-stat' or 'inc-stat', got {stat_type!r}"

    part_stat_filename = file_naming.get(stat_type, gzipped=False, partial=True)
    perm_stat_filename = file_naming.get(stat_type, gzipped=True)
    remote_stat_filename = file_naming.get(stat_type, encrypted=config.encryption, gzipped=config.compression)

    fh = dup_temp.get_fileobj_duppath(
        config.archive_dir_path, part_stat_filename, perm_stat_filename, remote_stat_filename, overwrite=True
    )
    return fh


def full_backup(col_stats):
    """
    Do full backup of directory to backend, using archive_dir_path.

    :param col_stats: The collection status.
    """
    if config.progress:
        progress.tracker = progress.ProgressTracker()
        # Fake a backup to compute total of moving bytes
        tarblock_iter = diffdir.DirFull(config.select)
        dummy_backup(tarblock_iter)
        # Store computed stats to compute progress later
        progress.tracker.set_evidence(diffdir.stats, True)
        # Reinit the config.select iterator, so
        # the core of duplicity can rescan the paths
        cli_main.set_selection()
        progress.progress_thread = progress.LogProgressThread()

    if config.dry_run:
        tarblock_iter = diffdir.DirFull(config.select)
        bytes_written = dummy_backup(tarblock_iter)
        col_stats.set_values(sig_chain_warning=None)
    else:
        sig_outfp = get_sig_fileobj("full-sig")
        man_outfp = get_man_fileobj("full")
        tarblock_iter = diffdir.DirFull_WriteSig(config.select, sig_outfp)
        bytes_written = write_multivol("full", tarblock_iter, man_outfp, sig_outfp, config.backend)

        # close sig file, send to remote, and rename to final
        sig_outfp.close()
        sig_outfp.to_remote()
        sig_outfp.to_final()

        # close manifest, send to remote, and rename to final
        man_outfp.close()
        man_outfp.to_remote()
        man_outfp.to_final()

        if config.progress:
            # Terminate the background thread now, if any
            progress.progress_thread.finished = True
            progress.progress_thread.join()
            progress.TransferProgress(
                100.0,
                0,
                progress.tracker.total_bytecount,
                progress.tracker.total_elapsed_seconds(),
                progress.tracker.speed,
                False,
            )

        col_stats.set_values(sig_chain_warning=None)

    write_json_stat("full-stat", bytes_written, col_stats)
    print_statistics(diffdir.stats, bytes_written)


def check_sig_chain(col_stats):
    """
    Get the last signature chain for incremental backup.

    :param col_stats: The collection status.
    :returns: The matched signature chain, or None if none available.
    """
    if not col_stats.matched_chain_pair:
        if config.action == "inc" and not config.implied_inc:
            raise errors.FatalError(
                _(
                    "Fatal Error: Unable to start incremental backup.  "
                    "Old signatures not found and incremental specified"
                ),
                log.ErrorCode.inc_without_sigs,
            )
        else:
            log.Warn(_("No signatures found, switching to full backup."))
        return None
    return col_stats.matched_chain_pair[0]


def print_statistics(stats, bytes_written):  # pylint: disable=unused-argument
    """
    If ``config.print_statistics`` is True, print stats after adding ``bytes_written``.

    :param stats: The statistics object.
    :param int bytes_written: The number of bytes written to the backend.
    """
    if config.print_statistics:
        diffdir.stats.TotalDestinationSizeChange = bytes_written
        logstring = diffdir.stats.get_stats_logstring(_("Backup Statistics"))
        log.Log(logstring, log.NOTICE, force_print=True)


def incremental_backup(sig_chain, col_stats=None):
    """
    Do incremental backup of directory to backend, using archive_dir_path.

    :param sig_chain: The signature chain to use.
    :param col_stats: The collection status.
    """
    if not config.restart:
        dup_time.setprevtime(sig_chain.end_time)
        if dup_time.curtime == dup_time.prevtime:
            time.sleep(2)
            dup_time.setcurtime()
            assert (
                dup_time.curtime != dup_time.prevtime
            ), "time not moving forward at appropriate pace - system clock issues?"

    if config.progress:
        progress.tracker = progress.ProgressTracker()
        # Fake a backup to compute total of moving bytes
        tarblock_iter = diffdir.DirDelta_WriteSig(config.select, sig_chain.get_fileobjs(), None)
        dummy_backup(tarblock_iter)
        # Store computed stats to compute progress later
        progress.tracker.set_evidence(diffdir.stats, False)
        # Reinit the config.select iterator, so
        # the core of duplicity can rescan the paths
        cli_main.set_selection()
        progress.progress_thread = progress.LogProgressThread()

    if config.dry_run:
        tarblock_iter = diffdir.DirDelta_WriteSig(config.select, sig_chain.get_fileobjs(), None)
        bytes_written = dummy_backup(tarblock_iter)
    else:
        new_sig_outfp = get_sig_fileobj("new-sig")
        new_man_outfp = get_man_fileobj("inc")
        tarblock_iter = diffdir.DirDelta_WriteSig(config.select, sig_chain.get_fileobjs(), new_sig_outfp)
        bytes_written = write_multivol("inc", tarblock_iter, new_man_outfp, new_sig_outfp, config.backend)

        if not config.skipped_inc:
            # write metadata to cache and remote
            # close sig file and rename to final
            new_sig_outfp.close()
            new_sig_outfp.to_remote()
            new_sig_outfp.to_final()

            # close manifest and rename to final
            new_man_outfp.close()
            new_man_outfp.to_remote()
            new_man_outfp.to_final()

        else:
            # drop metadata as no files as changed.
            new_sig_outfp.close()
            new_sig_outfp.clean_up()
            new_man_outfp.close()
            new_man_outfp.clean_up()

        if config.progress:
            # Terminate the background thread now, if any
            progress.progress_thread.finished = True
            progress.progress_thread.join()
            progress.TransferProgress(
                100.0,
                0,
                progress.tracker.total_bytecount,
                progress.tracker.total_elapsed_seconds(),
                progress.tracker.speed,
                False,
            )

    write_json_stat("inc-stat", bytes_written, col_stats)
    print_statistics(diffdir.stats, bytes_written)


def write_json_stat(stat_type, bytes_written, col_stats):
    """
    Write an extra statistic file if ``--jsonstat`` is specified.

    :param str stat_type: The name of the json_stat ("full-stat" or "inc-stat").
    :param int bytes_written: The number of bytes written in this run.
    :param col_stats: The collection status.
    """
    if config.jsonstat:
        json_stat = diffdir.stats.get_stats_json(col_stats)
        log.Log(json_stat, log.NOTICE, force_print=True)
        if not config.dry_run:
            stat_outfp = get_stat_fileobj(stat_type)
            diffdir.stats.TotalDestinationSizeChange = bytes_written
            stat_outfp.write(json_stat.encode())
            stat_outfp.close()
            if not config.skipped_inc:
                stat_outfp.to_remote()
                stat_outfp.to_final()
            else:
                stat_outfp.clean_up()


def list_current(col_stats):
    """
    List the files current in the archive (examining signatures only).

    :param col_stats: The collection status.
    """
    time = config.restore_time or dup_time.curtime
    sig_chain = col_stats.get_signature_chain_at_time(time)
    path_iter = diffdir.get_combined_path_iter(sig_chain.get_fileobjs(time))
    for path in path_iter:
        if path.difftype != "deleted":
            user_info = f"{dup_time.timetopretty(path.getmtime())} {os.fsdecode(path.get_relative_path())}"
            log_info = f"{dup_time.timetostring(path.getmtime())} {util.escape(path.get_relative_path())} {path.type}"
            log.Log(user_info, log.INFO, log.InfoCode.file_list, log_info, True)


def restore(col_stats):
    """
    Restore archive from the backend to the local path.

    :param col_stats: The collection status.
    """
    if config.dry_run:
        # Only prints list of required volumes when running dry
        restore_get_patched_rop_iter(col_stats)
        return
    if not patchdir.Write_ROPaths(config.local_path, restore_get_patched_rop_iter(col_stats)):
        if config.restore_path:
            raise errors.FatalError(
                _("%s not found in archive - no files restored.") % (os.fsdecode(config.restore_path)),
                log.ErrorCode.restore_path_not_found,
            )
        else:
            raise errors.FatalError(_("No files found in archive - nothing restored."), log.ErrorCode.no_restore_files)


def restore_get_patched_rop_iter(col_stats):
    """
    Return an iterator of patched ROPaths for the desired restore data.

    :param col_stats: The collection status.
    :returns: An iterator of patched ROPaths.
    :rtype: iterator
    """
    if config.restore_path:
        index = tuple(config.restore_path.split(b"/"))
    else:
        index = ()
    time = config.restore_time or dup_time.curtime
    backup_chain = col_stats.get_backup_chain_at_time(time)
    assert backup_chain, (
        f"No backup chain found for restore time {dup_time.timetostring(time)}; "
        f"available chains: {col_stats.all_backup_chains}"
    )
    backup_setlist = backup_chain.get_sets_at_time(time)
    restore_selection = None
    if config.action == "restore" and config.select_opts:
        restore_selection = selection.get_restore_selection(config.select_opts, config.select_files)
    num_vols = 0
    for s in backup_setlist:
        num_vols += len(s)
    cur_vol = [0]

    def get_fileobj_iter(backup_set):
        """
        Get file object iterator from backup_set contain given index
        """
        manifest = backup_set.get_manifest()
        volumes = manifest.get_containing_volumes(index)
        for vol_num in volumes:
            fobj = restore_get_enc_fileobj(
                backup_set.backend,
                backup_set.volume_name_dict[vol_num],
                manifest.volume_info_dict[vol_num],
            )
            if fobj is not None:
                yield fobj

            cur_vol[0] += 1
            log.Progress(_("Processed volume %d of %d") % (cur_vol[0], num_vols), cur_vol[0], num_vols)

    if hasattr(config.backend, "pre_process_download_batch") or config.dry_run:
        file_names = []
        for backup_set in backup_setlist:
            manifest = backup_set.get_manifest()
            volumes = manifest.get_containing_volumes(index)
            for vol_num in volumes:
                try:
                    file_names.append(backup_set.volume_name_dict[vol_num])
                except KeyError as e:
                    if config.ignore_errors:
                        log.Warn(f"difftar volume {vol_num} not found, ignoring per --ignore-errors.")
                    else:
                        raise errors.FatalError(
                            f"difftar volume {vol_num} not found in backup.\n"
                            f"Please check your archive and try again.\n"
                            f"Or use --ignore-errors to ignore this error.\n"
                            f"Error: {e}"
                        )
        if config.dry_run:
            log.Notice("Required volumes to restore:\n\t" + "\n\t".join(file_name.decode() for file_name in file_names))
            return None
        else:
            config.backend.pre_process_download_batch(file_names)

    fileobj_iters = list(map(get_fileobj_iter, backup_setlist))
    tarfiles = list(map(patchdir.TarFile_FromFileobjs, fileobj_iters))
    rop_iter = patchdir.tarfiles2rop_iter(tarfiles, index)
    if restore_selection:
        rop_iter = selection.filter_restore_path_iter(rop_iter, restore_selection)
    return rop_iter


def restore_get_enc_fileobj(backend, filename, volume_info):
    """
    Return plaintext fileobj from encrypted filename on backend

    If volume_info is set, the hash of the file will be checked,
    assuming some hash is available.  Also, if config.sign_key is
    set, a fatal error will be raised if file not signed by sign_key.

    with --ignore-errors set continue on hash mismatch

    """
    for n in range(1, config.num_retries + 1):
        """
        get the remote file
        """
        parseresults = file_naming.parse(filename)
        tdp = dup_temp.new_tempduppath(parseresults)
        backend.get(filename, tdp)

        """
        Verify hash of the remote file.
        """
        verified, hash_pair, calculated_hash = restore_check_hash(volume_info, tdp)
        if verified:
            break
        else:
            error_msg = "%s\n %s\n %s\n %s\n" % (
                _("Invalid data - %s hash mismatch for file:") % hash_pair[0],
                os.fsdecode(filename),
                _("Calculated hash: %s") % calculated_hash,
                _("Manifest hash: %s") % hash_pair[1],
            )
            log.Error(error_msg, code=log.ErrorCode.mismatched_hash)
    else:
        if config.ignore_errors:
            msg = f"Hash mismatch for: {os.fsdecode(filename)}"
            log.Warn(_("IGNORED_ERROR: WARNING: ignoring error as requested: %s: %s") % ("BadVolumeException", msg))
            # Do not try to actually read it as it is corrupted!
            return None
        else:
            raise errors.FatalError(error_msg, code=log.ErrorCode.mismatched_hash)
    fileobj = tdp.filtered_open_with_delete("rb")
    if parseresults.encrypted and config.gpg_profile.sign_key:
        restore_add_sig_check(fileobj)
    return fileobj


def restore_check_hash(volume_info, vol_path):
    """
    Check the hash of ``vol_path`` against the data in ``volume_info``.

    :param volume_info: The volume information containing the expected hash.
    :param Path vol_path: The local path of the volume to check.
    :returns: A tuple of (verified, expected_hash_pair, calculated_hash).
    :rtype: tuple
    """
    hash_pair = volume_info.get_best_hash()
    if hash_pair:
        calculated_hash = gpg.get_hash(hash_pair[0], vol_path)
        if calculated_hash != hash_pair[1]:
            return False, hash_pair, calculated_hash
    """
    Reached here, verification passed.
    """
    return True, hash_pair, calculated_hash


def restore_add_sig_check(fileobj):
    """
    Require a signature when closing ``fileobj`` that matches the GPG profile.

    :param fileobj: The file object to add the signature check to.
    """
    assert isinstance(fileobj, dup_temp.FileobjHooked) and isinstance(fileobj.fileobj, gpg.GPGFile), fileobj

    def check_signature():
        """
        Thunk run when closing volume file
        """
        actual_sig = fileobj.fileobj.get_signature()
        actual_sig = "None" if actual_sig is None else actual_sig
        sign_key = config.gpg_profile.sign_key
        sign_key = "None" if sign_key is None else sign_key
        ofs = -min(len(actual_sig), len(sign_key))
        if actual_sig[ofs:] != sign_key[ofs:]:
            raise errors.FatalError(
                _("Volume was signed by key %s, not %s") % (actual_sig[ofs:], sign_key[ofs:]),
                log.ErrorCode.unsigned_volume,
            )

    fileobj.addhook(check_signature)


def verify(col_stats):
    """
    Verify files, logging any differences found.

    :param col_stats: The collection status.
    """
    global exit_val
    collated = diffdir.collate2iters(restore_get_patched_rop_iter(col_stats), config.select)
    diff_count = 0
    total_count = 0
    for backup_ropath, current_path in collated:
        if not backup_ropath:
            backup_ropath = path.ROPath(current_path.index)
        if not current_path:
            current_path = path.ROPath(backup_ropath.index)
        if not backup_ropath.compare_verbose(current_path, config.compare_data):
            diff_count += 1
        total_count += 1
    # Unfortunately, ngettext doesn't handle multiple number variables, so we
    # split up the string.
    log.Notice(
        _("Verify complete: %s, %s.")
        % (_("%d file(s) compared") % total_count, _("%d difference(s) found") % diff_count)
    )
    if diff_count >= 1:
        exit_val = 1


def cleanup(col_stats):
    """
    Delete the extraneous files in the current backend.

    :param col_stats: The collection status.
    """
    ext_local, ext_remote = col_stats.get_extraneous()
    extraneous = ext_local + ext_remote
    if not extraneous:
        log.Warn(_("No extraneous files found, nothing deleted in cleanup."))
        return

    filestr = "\n".join(map(os.fsdecode, extraneous))
    if config.force:
        log.Notice(f"{_('Deleting these file(s) from backend:')}\n{filestr}")
        if not config.dry_run:
            col_stats.backend.delete(ext_remote)
            for fn in ext_local:
                try:
                    config.archive_dir_path.append(fn).delete()
                except Exception:
                    pass
    else:
        log.Notice(
            f"{_('Found the following file(s) to delete:')}\n{filestr}\n"
            "Run duplicity again with the --force option to actually delete."
        )


def maybe_delete_partial(col_stats, action):
    """
    Remove partial backups for cleanup and remove-* operations.

    :param col_stats: The collection status.
    :param str action: The action in progress.
    """

    # remove last partial backup and get new collection status
    if config.force:
        last_full_chain = col_stats.get_last_backup_chain()
        if last_full_chain:
            last_backup = last_full_chain.get_last()
            if last_backup.partial or not last_backup.is_complete() or last_backup.is_missing():
                log.Notice(_(f"Cleaning up previous partial backup set for {action}"))
                last_backup.delete()
                col_stats = dup_collections.CollectionsStatus(
                    config.backend,
                    config.archive_dir_path,
                ).set_values()
                log.Notice(_(f"Cleaned up previous partial backup at {col_stats.get_last_full_backup_time()}"))
    else:
        log.Notice(_(f"Need to use '{action} --force' to actually delete."))

    return col_stats


def remove_all_but_n_full(col_stats, action):
    """
    Remove backup files older than the last n full backups.

    :param col_stats: The collection status.
    :param str action: The action in progress.
    """
    assert config.keep_chains is not None, "--keep-chains must be set before removing old backups"

    col_stats = maybe_delete_partial(col_stats, action)

    config.remove_time = col_stats.get_nth_last_full_backup_time(config.keep_chains)

    remove_old(col_stats, action)


def remove_old(col_stats, action):
    """
    Remove backup files older than ``config.remove_time`` from the backend.

    :param col_stats: The collection status.
    :param str action: The action in progress.
    """

    def set_times_str(setlist):
        """
        Return string listing times of sets in setlist
        """
        return "\n".join([dup_time.timetopretty(s.get_time()) for s in setlist])

    def chain_times_str(chainlist):
        """
        Return string listing times of chains in chainlist
        """
        return "\n".join([dup_time.timetopretty(s.end_time) for s in chainlist])

    assert config.remove_time is not None, "config.remove_time must be set before removing old backups"

    col_stats = maybe_delete_partial(col_stats, action)

    req_list = col_stats.get_older_than_required(config.remove_time)
    if req_list:
        log.Warn(
            "%s\n%s\n%s"
            % (
                _("There are backup set(s) at time(s):"),
                set_times_str(req_list),
                _("Which can't be deleted because newer sets depend on them."),
            )
        )

    if col_stats.matched_chain_pair and col_stats.matched_chain_pair[1].end_time < config.remove_time:
        log.Warn(
            _(
                "Current active backup chain is older than specified time.  "
                "However, it will not be deleted.  To remove all your backups, "
                "manually purge the repository."
            )
        )

    chainlist = col_stats.get_chains_older_than(config.remove_time)

    if config.action == "remove-all-inc-of-but-n-full":
        # ignore chains without incremental backups:
        chainlist = list(
            x
            for x in chainlist
            if (isinstance(x, dup_collections.SignatureChain) and x.inclist)
            or (isinstance(x, dup_collections.BackupChain) and x.incset_list)
        )

    if not chainlist:
        log.Notice(_("No old backup sets found, nothing deleted."))
        return
    if config.force:
        log.Notice(_("Deleting backup chain(s) at time:") + "\n" + chain_times_str(chainlist))
        # Add signature files too, since they won't be needed anymore
        chainlist += col_stats.get_signature_chains_older_than(config.remove_time)
        chainlist.reverse()  # save oldest for last
        for chain in chainlist:
            # if action is remove_all_inc_of_but_n_full, remove only
            # incrementals one and not full
            if config.action == "remove-all-inc-of-but-n-full":
                if isinstance(chain, dup_collections.SignatureChain):
                    chain_desc = _("Deleting any incremental signature chain rooted at %s")
                else:
                    chain_desc = _("Deleting any incremental backup chain rooted at %s")
            else:
                if isinstance(chain, dup_collections.SignatureChain):
                    chain_desc = _("Deleting complete signature chain %s")
                else:
                    chain_desc = _("Deleting complete backup chain %s")
            log.Notice(chain_desc % dup_time.timetopretty(chain.end_time))
            if not config.dry_run:
                chain.delete(keep_full=(config.action == "remove-all-inc-of-but-n-full"))
        col_stats.set_values(sig_chain_warning=None)
    else:
        log.Notice(
            _("Found old backup chain(s) at the following time:")
            + "\n"
            + chain_times_str(chainlist)
            + "\n"
            + _("Rerun command with --force option to actually delete.")
        )


def sync_archive(col_stats):
    """
    Synchronize the local archive manifest file and signature chains.

    Copies missing files from the remote to the local archive as needed to
    ensure the local archive is synchronized with remote storage.

    :param col_stats: The collection status.
    """
    suffixes = [b".g", b".gpg", b".p7m", b".z", b".gz", b".part"]

    def is_needed(filename):
        """Indicates if the metadata file should be synced.

        In full sync mode, or if there's a collection misbehavior, all files
        are needed.

        Otherwise, only the metadata for the target chain needs sync.
        """
        if config.metadata_sync_mode == "full":
            return True
        assert config.metadata_sync_mode == "partial", "metadata_sync_mode must be 'partial' when not in full sync mode"
        parsed = file_naming.parse(filename)
        try:
            target_chain = col_stats.get_backup_chain_at_time(config.restore_time or dup_time.curtime)
        except dup_collections.CollectionsError:
            # With zero or multiple chains at this time, do a full sync
            return True
        if parsed.start_time is None and parsed.end_time is None:
            start_time = end_time = parsed.time
        else:
            start_time = parsed.start_time
            end_time = parsed.end_time

        return end_time >= target_chain.start_time and start_time <= target_chain.end_time

    def get_metafiles(filelist):
        """
        Return metafiles of interest from the file list.

        Files of interest are:
          - sigtar (signature files)
          - manifest (manifest files)
          - duplicity partial versions of the above

        Files excluded are:
          - non-duplicity files

        :param list filelist: A list of filenames.
        :returns: A dictionary of duplicity metadata files.
        :rtype: dict
        """
        metafiles = {}
        partials = {}
        need_passphrase = False
        for fn in filelist:
            pr = file_naming.parse(fn)
            if not pr:
                continue
            if pr.encrypted:
                need_passphrase = True
            if pr.type in ["full-sig", "new-sig"] or pr.manifest:
                base, ext = os.path.splitext(fn)
                if ext not in suffixes:
                    base = fn
                if pr.partial:
                    partials[base] = fn
                else:
                    metafiles[base] = fn
        return metafiles, partials, need_passphrase

    def copy_raw(src_iter, filename):
        """
        Copy data from src_iter to filename
        """
        file = open(filename, "wb")
        while True:
            try:
                data = src_iter.__next__().data
            except StopIteration:
                break
            file.write(data)
        file.close()

    def resolve_basename(fn):
        """
        Resolve the base name of a metadata file.

        :param str fn: The filename.
        :returns: A tuple of (parsed_result, local_name, remote_name).
        :rtype: tuple
        """
        pr = file_naming.parse(fn)

        base, ext = os.path.splitext(fn)
        if ext not in suffixes:
            base = fn

        suffix = file_naming.get_suffix(False, not pr.manifest)
        loc_name = base + suffix

        return pr, loc_name, fn

    def remove_local(fn):
        """
        Remove a local metadata file from the archive cache.
        """
        del_name = config.archive_dir_path.append(fn).name

        log.Notice(_("Deleting local %s (not authoritative at backend).") % os.fsdecode(del_name))
        try:
            util.ignore_missing(os.unlink, del_name)
        except Exception as e:
            log.Warn(_("Unable to delete %s: %s") % (os.fsdecode(del_name), util.uexc(e)))

    def copy_to_local(fn):
        """
        Copy remote file fn to local cache.
        """

        class Block(object):
            """
            Data block to return from SrcIter
            """

            def __init__(self, data):
                """
                Initialize a block with raw bytes.
                """
                self.data = data

        class SrcIter(object):
            """
            Iterate over source and return Block of data.
            """

            def __init__(self, fileobj):
                """
                Wrap a file object for block-based streaming.
                """
                self.fileobj = fileobj

            def __next__(self):
                """
                Read and return the next data block.
                """
                try:
                    res = Block(self.fileobj.read(self.get_read_size()))
                except Exception as e:
                    if hasattr(self.fileobj, "name"):
                        name = self.fileobj.name
                        # name may be a path
                        if hasattr(name, "name"):
                            name = name.name
                    else:
                        name = None
                    raise errors.FatalError(
                        _("Failed to read %s: %s") % (os.fsdecode(fn), util.uexc(e)), log.ErrorCode.generic
                    )
                if not res.data:
                    self.fileobj.close()
                    raise StopIteration
                return res

            def get_read_size(self):
                """
                Return the per-read chunk size in bytes.
                """
                return 128 * 1024

            def get_footer(self):
                """
                Return trailing footer bytes for stream consumers.
                """
                return b""

        log.Notice(_("Copying %s to local cache.") % os.fsdecode(fn))

        pr, loc_name, rem_name = resolve_basename(fn)

        fileobj = config.backend.get_fileobj_read(fn)
        src_iter = SrcIter(fileobj)
        tdp = dup_temp.new_tempduppath(file_naming.parse(loc_name))
        if pr.manifest:
            copy_raw(src_iter, tdp.name)
        else:
            gpg.GzipWriteFile(src_iter, tdp.name, size=sys.maxsize)
        tdp.setdata()
        tdp.move(config.archive_dir_path.append(loc_name))

    # skip if remote access is not configured
    if not config.check_remote:
        return

    # get remote metafile list
    remlist = config.backend.list()
    remote_metafiles, ignored, rem_needpass = get_metafiles(remlist)

    # get local metafile list
    loclist = config.archive_dir_path.listdir()
    local_metafiles, local_partials, loc_needpass = get_metafiles(loclist)

    # we have the list of metafiles on both sides. remote is always
    # authoritative. figure out which are local spurious (should not
    # be there) and missing (should be there but are not).
    local_keys = list(local_metafiles.keys())
    remote_keys = list(remote_metafiles.keys())

    local_missing = []
    local_spurious = []

    for key in remote_keys:
        # If we lost our cache, re-get the remote file.  But don't do it if we
        # already have a local partial.  The local partial will already be
        # complete in this case (seems we got interrupted before we could move
        # it to its final location).
        if key not in local_keys and key not in local_partials and is_needed(key):
            local_missing.append(remote_metafiles[key])

    for key in local_keys:
        # If we have a file locally that is unnecessary, delete it.  Also
        # delete final versions of partial files because if we have both, it
        # means the write of the final version got interrupted.
        if key not in remote_keys or key in local_partials:
            local_spurious.append(local_metafiles[key])

    # finally finish the process
    if not local_missing and not local_spurious:
        log.Notice(_("Local and Remote metadata are synchronized, no sync needed."))
    else:
        local_missing.sort()
        local_spurious.sort()
        if not config.dry_run:
            log.Notice(_("Synchronizing remote metadata to local cache..."))
            if local_missing and (rem_needpass or loc_needpass):
                # password for the --encrypt-key
                config.gpg_profile.passphrase = get_passphrase(1, "sync")
            for fn in local_spurious:
                remove_local(fn)
            if hasattr(config.backend, "pre_process_download_batch"):
                config.backend.pre_process_download_batch(local_missing)
            for fn in local_missing:
                copy_to_local(fn)
            col_stats.set_values()
        else:
            if local_missing:
                log.Notice(
                    _("Sync would copy the following from remote to local:")
                    + "\n"
                    + "\n".join(map(os.fsdecode, local_missing))
                )
            if local_spurious:
                log.Notice(
                    _("Sync would remove the following spurious local files:")
                    + "\n"
                    + "\n".join(map(os.fsdecode, local_spurious))
                )


def check_last_manifest(col_stats):
    """
    Check the consistency and hostname/directory of the last manifest.

    :param col_stats: The collection status.
    :raises AssertionError: If no backup chains are found.
    """
    assert col_stats.all_backup_chains, "No backup chains found; cannot check last manifest without any chains present"
    last_backup_set = col_stats.all_backup_chains[-1].get_last()
    # check remote manifest only if we can decrypt it (see #1729796)
    last_backup_set.check_manifests(check_remote=config.check_remote)


def check_resources(action):
    """
    Check for sufficient resources to run the operation.

    Checks:
      - temp space for volume build
      - enough max open files

    Raises a fatal error if resources are not sufficient.

    :param str action: The action in progress.
    """
    if action in ["full", "inc", "restore"]:
        # Make sure we have enough resouces to run
        # First check disk space in temp area.
        tempfile, tempname = tempdir.default().mkstemp()
        os.close(tempfile)
        # strip off the temp dir and file
        tempfs = os.path.sep.join(tempname.split(os.path.sep)[:-2])
        try:
            stats = os.statvfs(tempfs)
        except Exception:
            raise errors.FatalError(_("Unable to get free space on temp."), log.ErrorCode.get_freespace_failed)
        # Calculate space we need for at least 2 volumes of full or inc
        # plus about 30% of one volume for the signature files.
        freespace = stats.f_frsize * stats.f_bavail
        needspace = ((config.concurrency + 2) * config.volsize) + int(0.30 * config.volsize)
        if freespace < needspace:
            raise errors.FatalError(
                _(f"Temp space has {freespace:,} available, backup needs approx {needspace:,}."),
                log.ErrorCode.not_enough_freespace,
            )
        else:
            log.Info(_(f"Temp has {freespace:,} available, backup will use approx {needspace:,}."))

        # Some environments like Cygwin run with an artificially
        # low value for max open files.  Check for safe number.
        try:
            soft, hard = resource.getrlimit(resource.RLIMIT_NOFILE)
        except resource.error:
            raise errors.FatalError(_("Unable to get max open files."), log.ErrorCode.get_ulimit_failed)
        maxopen = min([l for l in (soft, hard) if l > -1])
        if maxopen < 1024:
            raise errors.FatalError(
                _(
                    f"Max open files of {maxopen} is too low, should be >= 1024.\n"
                    f"Use 'ulimit -n 1024' or higher to correct.\n"
                ),
                log.ErrorCode.maxopen_too_low,
            )


def log_startup_parms(verbosity=log.INFO):
    """
    log Python, duplicity, and system versions
    """
    log.Log("=" * 80, verbosity)
    log.Log(f"duplicity {__version__}", verbosity)
    cmd_line = " ".join([os.fsdecode(arg) for arg in sys.argv])
    log.Log(f"Args: {backend.Backend.munge_password(cmd_line)}", verbosity)
    log.Log(" ".join(platform.uname()), verbosity)
    log.Log(f"{sys.executable or sys.platform} {sys.version}", verbosity)
    log.Log("=" * 80, verbosity)


class Restart(object):
    """
    Class to aid in restart of inc or full backup.
    Instance in config.restart if restart in progress.
    """

    def __init__(self, last_backup):
        """
        Initialize restart state from the last partial backup set.
        """
        self.time = None
        self.type = None
        self.start_time = None
        self.end_time = None
        self.start_vol = None
        self.last_index = None
        self.last_block = None
        self.last_backup = last_backup
        self.setParms(last_backup)

    def setParms(self, last_backup):
        """
        Derive restart type, times, and starting volume metadata.
        """
        if last_backup.time:
            self.type = "full"
            self.time = last_backup.time
        else:
            self.type = "inc"
            self.end_time = last_backup.end_time
            self.start_time = last_backup.start_time
        # We start one volume back in case we weren't able to finish writing
        # the most recent block.  Actually checking if we did (via hash) would
        # involve downloading the block.  Easier to just redo one block.
        self.start_vol = max(last_backup.get_last_continuous_volume() - 1, 0)

    def checkManifest(self, mf):
        """
        Validate manifest state and repair or restart when inconsistent.
        """
        mf_len = len(mf.volume_info_dict)
        if (mf_len != self.start_vol) or not (mf_len and self.start_vol):
            if self.start_vol == 0:
                # upload of 1st vol failed, clean and restart
                log.Notice(
                    _(
                        "RESTART: The first volume failed to upload before termination.\n"
                        "         Restart is impossible...starting backup from beginning."
                    )
                )
                self.last_backup.delete()
                os.execve(sys.argv[0], sys.argv, os.environ)
            elif mf_len - self.start_vol > 0:
                # upload of N vols failed, fix manifest and restart
                log.Notice(
                    _(
                        "RESTART: Volumes %d to %d failed to upload before termination.\n"
                        "         Restarting backup at volume %d."
                    )
                    % (self.start_vol + 1, mf_len, self.start_vol + 1)
                )
                for vol in range(self.start_vol + 1, mf_len + 1):
                    mf.del_volume_info(vol)
            else:
                # this is an 'impossible' state, remove last partial and restart
                log.Notice(
                    _(
                        "RESTART: Impossible backup state: manifest has %d vols, remote has %d vols.\n"
                        "         Restart is impossible ... duplicity will clean off the last partial\n"
                        "         backup then restart the backup from the beginning."
                    )
                    % (mf_len, self.start_vol)
                )
                self.last_backup.delete()
                os.execve(sys.argv[0], sys.argv, os.environ)

    def setLastSaved(self, mf):
        """
        Record the last saved index and block from the manifest.
        """
        vi = mf.volume_info_dict[self.start_vol]
        self.last_index = vi.end_index
        self.last_block = vi.end_block or 0


def main():
    """
    Start/end here
    """
    # per bug https://bugs.launchpad.net/duplicity/+bug/931175
    # duplicity crashes when python optimization is set, so check
    # and refuse to run if it is set.
    if sys.flags.optimize:
        raise errors.FatalError(
            dedent(_("""\
                    Setting python optimization causes duplicity to fail to
                    recognize its own backups.  Please remove PYTHONOPTIMIZE
                    from the environment or -O and -OO from the commandline,
                    then rerun the backup.

                    See https://bugs.launchpad.net/duplicity/+bug/931175
                    """)),
            log.ErrorCode.pythonoptimize_set,
        )

    # if python is run setuid, it's only partway set,
    # so make sure to run with euid/egid of root
    if os.geteuid() == 0:
        # make sure uid/gid match euid/egid
        os.setuid(os.geteuid())
        os.setgid(os.getegid())

    # set the current time strings (make it available for command line processing)
    dup_time.setcurtime()

    # determine what action we're performing and process command line
    action = cli_main.process_command_line(sys.argv[1:])

    # log some status info
    log_startup_parms(log.INFO)

    # make sure we have lock
    util.acquire_lockfile()

    try:
        # perform action
        do_backup(action)
    finally:
        # release lock
        util.release_lockfile()
        log.Notice("Duplicity backup complete.")


def do_backup(action):
    """
    Execute the selected action after preparing state and prerequisites.
    """
    # set the current time strings again now that we have time separator
    if config.current_time:
        dup_time.setcurtime(config.current_time)
    else:
        dup_time.setcurtime()

    # check for disk space and available file handles
    check_resources(action)

    # get current collection status
    col_stats = dup_collections.CollectionsStatus(
        config.backend,
        config.archive_dir_path,
        first=True,
    ).set_values()

    # check archive synch with remote, fix if needed
    if action not in skips_sync_archive:
        sync_archive(col_stats)

    while True:
        # if we have to clean up the last partial, then col_stats are invalidated
        # and we have to start the process all over again until clean.
        if action in ["full", "inc"]:
            last_full_chain = col_stats.get_last_backup_chain()
            if not last_full_chain:
                break
            last_backup = last_full_chain.get_last()
            if last_backup.partial:
                # set restart parms from last_backup info
                config.restart = Restart(last_backup)
                # (possibly) reset action
                action = config.restart.type
                # reset the time strings
                if action == "full":
                    dup_time.setcurtime(config.restart.time)
                else:
                    dup_time.setcurtime(config.restart.end_time)
                    dup_time.setprevtime(config.restart.start_time)
                # log it -- main restart heavy lifting is done in write_multivol
                log.Notice(_(f"Last {action} backup left a partial set, restarting."))
                break
            elif not last_backup.is_complete() or last_backup.is_missing():
                # remove last partial backup and get new collection status
                log.Notice(_(f"Cleaning up previous partial {action} backup set, restarting."))
                last_backup.delete()
                col_stats = dup_collections.CollectionsStatus(
                    config.backend,
                    config.archive_dir_path,
                ).set_values()
                continue
            else:
                break
        break

    # OK, now we have a stable collection
    last_full_time = col_stats.get_last_full_backup_time()
    last_chain = col_stats.get_last_backup_chain()
    if last_full_time > 0:
        log.Notice(f"{_('Last full backup date:')} {dup_time.timetopretty(last_full_time)}")
    else:
        log.Notice(_("Last full backup date: none"))
    if (
        not config.restart
        and action in ["inc"]
        and config.full_if_older_than is not None
        and last_full_time < dup_time.curtime - config.full_if_older_than
    ):
        log.Notice(_("Last full backup is too old, forcing full backup"))
        action = "full"
    if (
        not config.restart
        and action in ["inc"]
        and config.full_if_n_inc is not None
        and last_chain is not None
        and len(last_chain.incset_list) >= config.full_if_n_inc
    ):
        log.Notice(_("Latest chain has too many incrementals, forcing full backup"))
        action = "full"
    dup_collections.PrintCollectionStatus(col_stats)

    # get the passphrase if we need to based on action/options
    config.gpg_profile.passphrase = get_passphrase(1, action)

    if action == "restore":
        restore(col_stats)
    elif action == "verify":
        verify(col_stats)
    elif action == "list-current-files":
        list_current(col_stats)
    elif action == "collection-status":
        if config.show_changes_in_set is not None:
            if not config.jsonstat:
                # print classic stats
                dup_collections.PrintCollectionChangesInSet(col_stats, config.show_changes_in_set, True)
            else:
                # print json stat
                json_stat = col_stats.get_changes_in_set_json(config.show_changes_in_set)
                log.Log(str(json_stat), 8, log.InfoCode.collection_status, None, True)
        elif not config.file_changed:
            dup_collections.PrintCollectionStatus(col_stats, True)
        else:
            dup_collections.PrintCollectionFileChangedStatus(col_stats, config.file_changed, True)
    elif action == "cleanup":
        cleanup(col_stats)
    elif action == "remove-older-than":
        remove_old(col_stats, action)
    elif action in ["remove-all-but-n-full", "remove-all-inc-of-but-n-full"]:
        remove_all_but_n_full(col_stats, action)
    elif action == "sync":
        sync_archive(col_stats)
    else:
        assert action in ["full", "inc"], action
        # the passphrase for full and inc is used by --sign-key
        # the sign key can have a different passphrase than the encrypt
        # key, therefore request a passphrase
        if config.gpg_profile.sign_key:
            config.gpg_profile.signing_passphrase = get_passphrase(1, action, True)

        # if there are no recipients (no --encrypt-key), it must be a
        # symmetric key. Therefore, confirm the passphrase
        if not (config.gpg_profile.recipients or config.gpg_profile.hidden_recipients):
            config.gpg_profile.passphrase = get_passphrase(2, action)
            # a limitation in the GPG implementation does not allow for
            # inputting different passphrases, this affects symmetric+sign.
            # Allow an empty passphrase for the key though to allow a non-empty
            # symmetric key
            if (
                config.gpg_profile.signing_passphrase
                and config.gpg_profile.passphrase != config.gpg_profile.signing_passphrase
            ):
                raise errors.FatalError(
                    _(
                        "When using symmetric encryption, the signing passphrase "
                        "must equal the encryption passphrase."
                    ),
                    log.ErrorCode.user_error,
                )

        if action == "full":
            full_backup(col_stats)
        else:  # attempt incremental
            sig_chain = check_sig_chain(col_stats)
            # action == "inc" was requested, but no full backup is available
            if not sig_chain:
                full_backup(col_stats)
            else:
                if not config.restart:
                    # only ask for a passphrase if there was a previous backup
                    if col_stats.all_backup_chains:
                        config.gpg_profile.passphrase = get_passphrase(1, action)
                        check_last_manifest(col_stats)  # not needed for full backups
                incremental_backup(sig_chain, col_stats)

        # if action in ["full", "inc"] and config.check_remote:
        #     dup_collections.CollectionsStatus(
        #         config.backend,
        #         config.archive_dir_path,
        #     ).set_values()

    config.backend.close()
    log.shutdown()
    if exit_val is not None:
        sys.exit(exit_val)
